"""
Wrappers around neuroglancer classes that can compute metadata and
optional attributes that neuroglancer typically delegates to the frontend.

Dependencies
------------
* nifti://          -> nibabel
* zarr://           -> zarr
* precomputed://    -> cloudvolume
"""
# stdlib
import functools
import json
import logging
import time
from itertools import product
from os import PathLike
from types import EllipsisType
from typing import Callable, Iterator

# externals
import nibabel as nib
import numpy as np
from nibabel.spatialimages import HeaderDataError
from numpy.typing import ArrayLike
from upath import UPath

# internals
import ngtools.transforms as T
import ngtools.units as U
from ngtools.opener import (
    exists,
    filesystem,
    open,  # !!! hides builtin open
    parse_protocols,
    read_json,
)
from ngtools.optionals import try_from_import, try_import, try_import_as
from ngtools.utils import Wraps

# optionals
ng = try_import_as('neuroglancer', fallback='ngtools._nglite')
zarr = try_import('zarr.storage')
CloudVolume = try_from_import('cloudvolume', 'CloudVolume')

LOG = logging.getLogger(__name__)


class _SizedCache(dict):
    def __init__(self, max_size: int | None = None) -> None:
        self._max_size = max_size
        self._keys = []

    def __setitem__(self, key: object, value: object) -> None:
        if key in self:
            del self[key]
        super().__setitem__(key, value)
        if self._max_size:
            keys = list(self.keys())
            while len(self) > self._max_size:
                del self[keys.pop(0)]


_DATASOURCE_REGISTRY = {}
_DATASOURCE_INFO_CACHE = _SizedCache(128)


def _quantiles(q: float | list[float], data: ArrayLike) -> np.ndarray[float]:
    """Compute (efficiently) intensity quantiles."""
    tic = time.time()
    steps = [max(1, x//64) for x in data.shape]
    slicer = (Ellipsis,) + tuple(slice(None, None, step) for step in steps)
    data = np.asarray(data[slicer])
    data = data[np.isfinite(data) & (data != 0)]
    quantiles = np.quantile(data, q)
    toc = time.time()
    LOG.info(f"Compute quantiles: {toc-tic} s")
    return quantiles


def datasource(formats: str | list[str]) -> Callable[[type], type]:
    """
    Register a format-specific datasource class.

    Parameters
    ----------
    formats : str | list[str]
        Formats that should fallback to the decorated class.
    """
    if isinstance(formats, str):
        formats = [formats]

    def decorator(cls: type[LayerDataSource]) -> type[LayerDataSource]:
        for format in formats:
            _DATASOURCE_REGISTRY[format] = cls
        cls.PROTOCOLS = list(formats)
        return cls

    return decorator


SourceType = (
    dict |                                          # json
    str | PathLike |                                # uri
    ng.LayerDataSource |                            # data source
    ng.LocalVolume | ng.skeleton.SkeletonSource     # local source
)


class LayerDataSources(Wraps(ng.LayerDataSources)):
    """List of data sources."""

    _DataSourceLike = (
        str |
        ng.local_volume.LocalVolume |
        ng.skeleton.SkeletonSource |
        ng.LayerDataSource
    )
    _DataSourcesLike = (
        _DataSourceLike |
        list[_DataSourceLike | dict | None] |
        ng.LayerDataSources |
        dict |
        None
    )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Trigger data source computation so that defaults that differ
        # between ngtools and neuroglancer are set according to ngtools.
        for i, val in enumerate(self):
            self[i] = val

    @functools.wraps(ng.LayerDataSources.__iter__)
    def __iter__(self) -> Iterator[ng.LayerDataSource]:
        for source in super().__iter__():
            yield LayerDataSource(source)

    @functools.wraps(ng.LayerDataSources.__getitem__)
    def __getitem__(self, *args, **kwargs) -> "LayerDataSource":
        return LayerDataSource(super().__getitem__(*args, **kwargs))

    @functools.wraps(ng.LayerDataSources.pop)
    def pop(self, *args, **kwargs) -> "LayerDataSource":  # noqa: D102
        return LayerDataSource(super().pop(*args, **kwargs))


class _LayerDataSourceFactory(type):
    """Factory for LayerDataSource objects."""

    _LocalSource = (ng.local_volume.LocalVolume, ng.skeleton.SkeletonSource)
    _DataSourceLike = (
        str |
        ng.local_volume.LocalVolume |
        ng.skeleton.SkeletonSource |
        ng.LayerDataSource |
        dict |
        None
    )

    def __call__(
        cls,
        arg: _DataSourceLike = None,
        *args,
        **kwargs
    ) -> "LayerDataSource":
        # Only use the factory if it is not called from a subclass
        if cls is not LayerDataSource:
            obj = super().__call__(arg, *args, **kwargs)
            return obj

        # Get url
        if kwargs.get("url", ""):
            url = kwargs["url"]
            if isinstance(url, (str, PathLike)):
                url = kwargs["url"] = str(url)
        elif isinstance(arg, (str, PathLike)):
            url = arg = str(arg)
        elif hasattr(arg, "url"):
            url = arg.url
        elif isinstance(arg, dict) and "url" in arg:
            url = arg["url"]
        elif not isinstance(arg, cls._LocalSource):
            raise ValueError("Missing data source url")

        # If local object -> delegate
        if isinstance(url, ng.LocalVolume):
            return LocalVolumeDataSource(arg, *args, **kwargs)
        if isinstance(url, ng.skeleton.SkeletonSource):
            return LocalSkeletonDataSource(arg, *args, **kwargs)

        parsed_url = parse_protocols(url)

        # If python:// url -> local object, but retrieved from the viewer.
        if parsed_url.stream == "python":

            # Get local object
            path = parsed_url.url
            source_type = path.split("://")[-1].strip("/").split("/")[0]
            kls = {
                "volume": LocalVolumeDataSource,
                "skeleton": LocalSkeletonDataSource
            }[source_type]

            # Defer to LocalDataSource factory
            # TODO: build a `LocalDataSourceFactory` to defer even more?
            return kls(arg, *args, **kwargs)

        # If local:// url -> local annotations
        if parsed_url.url == "local://annotations":
            return LocalAnnotationDataSource(arg, *args, **kwargs)

        # If format protocol provided, use it
        if parsed_url.format in _DATASOURCE_REGISTRY:
            format = parsed_url.format
            LOG.debug(f"LayerDataSource - use format hint: {format}")
            return _DATASOURCE_REGISTRY[format](arg, *args, **kwargs)

        # Otherwise, check for extensions
        for format, kls in _DATASOURCE_REGISTRY.items():
            if parsed_url.url.endswith((format, format+".gz", format+".bz2")):
                LOG.debug(f"LayerDataSource - found extension: {format}")
                try:
                    obj = kls(arg, *args, **kwargs)
                    LOG.debug(f"LayerDataSource - {format} (success)")
                    return obj
                except Exception:
                    continue

        if False:  # Do not use for now -- need to implement sniffers
            # Otherwise, try all
            for format, kls in _DATASOURCE_REGISTRY.items():
                LOG.debug(f"LayerDataSource - try format: {format}")
                try:
                    obj = kls(arg, *args, **kwargs)
                    LOG.debug(f"LayerDataSource - {format} (success)")
                    return obj
                except Exception:
                    continue

        # Otherwise. build a simple source
        LOG.debug("LayerDataSource - Fallback to simple LayerDataSource")
        return super().__call__(arg, *args, **kwargs)


class DataSourceInfo:
    """Base class for source-specific metadata."""

    ...


class LayerDataSource(Wraps(ng.LayerDataSource),
                      metaclass=_LayerDataSourceFactory):
    """A wrapper around `ng.LayerDataSource` that computes metadata."""

    PROTOCOLS: list[str] = []

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # ensure that the url has the right format
        if isinstance(self.url, str):
            url = parse_protocols(self.url).url
            if self.PROTOCOLS:
                url = self.PROTOCOLS[0] + "://" + url
            self.url = url

    def _compute_info(self) -> DataSourceInfo:
        """Compute format-specific metadata."""
        ...

    def to_json(self) -> dict:  # noqa: D102
        # ensure the transform that we guess is saved in the ng object
        self.transform = self.transform
        return super().to_json()

    @property
    def info(self) -> DataSourceInfo:
        """Lazy access to format-specific metadata."""
        if not getattr(self, '_info', None):
            LOG.debug("Trigger _compute_info")
            url = self.local_url
            if url:
                url = parse_protocols(url).url
            key = (type(self), url)
            if key not in _DATASOURCE_INFO_CACHE:
                LOG.debug("Recompute info")
                try:
                    _DATASOURCE_INFO_CACHE[key] = self._compute_info()
                except Exception:
                    LOG.warning(f"Could not compute info for: {key}")
            else:
                LOG.debug("Use cached info")
            self._info = _DATASOURCE_INFO_CACHE.get(key, None)
        return self._info

    @classmethod
    def _raise_not_implemented_error(cls, name: str) -> None:
        raise NotImplementedError(
            f"`{name}` not implemented for this format ({cls.__name__})."
        )

    def __set__url__(self, value: str) -> str:
        LocalObject = (ng.local_volume.LocalVolume, ng.skeleton.SkeletonSource)
        if not isinstance(value, LocalObject):
            format = parse_protocols(value)[1]
            if format not in self.PROTOCOLS:
                raise ValueError(
                    "Cannot assign uri with protocol", format, "in "
                    "data source of type:", type(self)
                )
            self._info = None
        return value

    @property
    def local_url(self) -> str:
        """Path to url, even if file is local."""
        url = self.url
        if isinstance(url, (ng.LocalVolume, ng.skeleton.SkeletonSource)):
            if hasattr(self, '_url'):
                return self._url
            if hasattr(url, 'url'):
                return url.url
            return None
        return url

    @property
    def __default_transform__(self) -> ng.CoordinateSpaceTransform:
        return getattr(self, '_transform', None)

    def __get_transform__(self) -> ng.CoordinateSpaceTransform:
        """
        If a transform has been explicitely set, return it,
        else, return the transform implicitely defined by neuroglancer.
        """
        assigned_trf = self._wrapped.transform
        if (
            assigned_trf is None or
            assigned_trf.matrix is None or
            assigned_trf.input_dimensions is None or
            assigned_trf.output_dimensions is None
        ):
            tic = time.time()
            implicit_trf = getattr(self, '_transform', None)
            if implicit_trf is None:
                return assigned_trf
            if assigned_trf is None:
                return implicit_trf
            toc = time.time()
            LOG.debug(
                f"{type(self).__name__}: "
                "Compute implicit transform: "
                f"{toc - tic} s"
            )
            matrix = assigned_trf.matrix
            if matrix is None:
                matrix = implicit_trf.matrix
            idims = (
                assigned_trf.input_dimensions or
                implicit_trf.input_dimensions
            )
            odims = (
                assigned_trf.output_dimensions or
                implicit_trf.input_dimensions
            )
            self._wrapped.transform = ng.CoordinateSpaceTransform(
                matrix=matrix,
                input_dimensions=idims,
                output_dimensions=odims,
            )
            return self._wrapped.transform
        else:
            return assigned_trf

    @property
    def shape(self) -> list[int]:
        """Return the shape of the array (spatial dimensions first)."""
        if hasattr(self, "_shape"):
            return self._shape
        self._raise_not_implemented_error("shape")

    @property
    def rank(self) -> int:
        """Return the number of dimensions in the array."""
        return len(self.shape)

    ndim = rank

    @property
    def dtype(self) -> np.dtype:
        """Return the data type of the array."""
        if hasattr(self, "_dtype"):
            return self._dtype
        self._raise_not_implemented_error("dtype")

    @property
    def slope_inter(self) -> tuple[float, float]:
        """Affine scaling to apply to the stored intensity."""
        if hasattr(self, "_slope_inter"):
            return self._slope_inter
        return (1.0, 0.0)

    @property
    def dataobj(self) -> ArrayLike:
        """Return an array-like object pointing to the data."""
        return self.get_dataobj(0)

    @property
    def nb_levels(self) -> int:
        """Number of pyramid levels."""
        if hasattr(self, "_nb_levels"):
            return self._nb_levels
        return 1

    def get_dataobj(self, level: int = 0, mode: str = "r") -> ArrayLike:
        """Return an array-like object pointing to a pyramid level."""
        if level < 0:
            level = self.nb_levels + level
        return self._get_dataobj(level, mode)

    def _get_dataobj(self, level: int = 0, mode: str = "r") -> ArrayLike:
        # Implementation classes should implement this method.
        self._raise_not_implemented_error("get_dataobj")

    _IndexType = int | bool | slice | ArrayLike | EllipsisType
    _SlicerType = _IndexType | tuple[_IndexType]

    def __getitem__(self, slicer: _SlicerType) -> ArrayLike:
        return self.dataobj[slicer]

    def quantiles(self, q: float | list[float]) -> np.ndarray:
        """Compute data quantiles."""
        spatial_dims = self.transform.input_dimensions
        spatial_dims = [
            i for i, unit in enumerate(spatial_dims.units)
            if unit.endswith("m")
        ]
        for i in range(self.nb_levels):
            dat = self.get_dataobj(i, mode="r")
            if all([dat.shape[i] <= 64 for i in spatial_dims]):
                break
        return _quantiles(q, dat)

    def apply_transform(self, *args: ng.CoordinateSpaceTransform
                        ) -> "LayerDataSource":
        """Apply an additional transform in model space."""
        self.transform = T.compose(*args, self.transform)
        return self

    @property
    def input_dimensions(self) -> ng.CoordinateSpace:
        """Input dimensions."""
        return self.transform.input_dimensions

    inputDimensions = input_dimensions

    @property
    def output_dimensions(self) -> ng.CoordinateSpace:
        """Output dimensions."""
        return self.transform.output_dimensions

    outputDimensions = output_dimensions

    @property
    def input_bbox(self) -> list[list[float]]:
        """Bounding box, in input_dimensions space and units."""
        shape = self.shape
        # FIXME: should there be a half voxel shift?
        min = [0.0] * self.rank
        max = shape
        return [min, max]

    @property
    def output_bbox(self) -> list[list[float]]:
        """Bounding box, in output_dimensions space and units."""
        bbox = self.input_bbox

        mn = np.full([self.rank], +float('inf'))
        mx = np.full([self.rank], -float('inf'))

        for corner in product([0, 1], repeat=self.rank):
            coord = [bbox[j][i] for i, j in enumerate(corner)]

            mat = np.eye(self.rank+1)[:-1]
            mat[:, -1] = coord
            coord = ng.CoordinateSpaceTransform(
                input_dimensions=self.input_dimensions,
                output_dimensions=self.input_dimensions,
                matrix=mat,
            )

            coord = T.compose(self.transform, coord)
            coord = coord.matrix[:, -1]

            mn = np.minimum(mn, coord)
            mx = np.maximum(mx, coord)

        bbox = [mn.tolist(), mx.tolist()]
        return bbox

    @property
    def input_center(self) -> list[float]:
        """Center of the field of view in input dimensions space and units."""
        bbox = np.asarray(self.input_bbox)
        return ((bbox[0] + bbox[1]) / 2).tolist()

    @property
    def output_center(self) -> list[float]:
        """Center of the field of view in output dimensions space and units."""
        center = self.input_center
        mat = np.eye(self.rank+1)[:-1]
        mat[:, -1] = center
        center = ng.CoordinateSpaceTransform(
            input_dimensions=self.input_dimensions,
            output_dimensions=self.input_dimensions,
            matrix=mat,
        )
        center = T.compose(self.transform, center)
        center = center.matrix[:, -1]
        return center

    @property
    def input_bbox_size(self) -> list[float]:
        """Center of the field of view in input dimensions space and units."""
        bbox = np.asarray(self.input_bbox)
        return (bbox[1] - bbox[0]).tolist()

    @property
    def output_bbox_size(self) -> list[float]:
        """Center of the field of view in output dimensions space and units."""
        bbox = np.asarray(self.output_bbox)
        return (bbox[1] - bbox[0]).tolist()

    @property
    def output_voxel_size(self) -> list[float]:
        """Voxel size in model space."""
        mat = np.eye(self.rank+1)[:-1]
        mat[:, -1] = 1
        size = ng.CoordinateSpaceTransform(
            input_dimensions=self.input_dimensions,
            output_dimensions=self.input_dimensions,
            matrix=mat,
        )
        size = T.compose(self.transform, size)
        size = size.matrix[range(self.rank), range(self.rank)]
        return size.tolist()


class VolumeInfo(DataSourceInfo):
    """Base class for volume metadata."""

    def __init__(self, url: str) -> None:
        """
        Parameters
        ----------
        url : str
            URL to the file
        """
        self.url = str(url)

    def getNbLevels(self) -> int:
        """Compute number of dimensions."""
        return 1

    def getShape(self, level: int = 0) -> list[int]:
        """Compute array shape."""
        raise NotImplementedError

    def getRank(self) -> int:
        """Compute number of dimensions."""
        return len(self.getShape())

    def getSlopeInter(self) -> tuple[float, float]:
        """Affine intensity scaling."""
        return (1.0, 0.0)

    def getInputNames(self) -> list[str]:
        """Compute names of input axes."""
        raise NotImplementedError

    def getOutputNames(self) -> list[str]:
        """Compute names of output axes."""
        raise NotImplementedError

    def getInputUnits(self) -> list[str]:
        """Compute all units."""
        raise NotImplementedError

    def getOutputUnits(self) -> list[str]:
        """Compute all units."""
        raise NotImplementedError

    def getInputScales(self) -> list[float]:
        """Compute input scales."""
        raise NotImplementedError

    def getOutputScales(self) -> list[float]:
        """Compute output scales."""
        raise NotImplementedError

    def getDataType(self) -> np.dtype:
        """Compute numpy data type."""
        raise NotImplementedError

    def getInputDimensions(self) -> ng.CoordinateSpace:
        """Compute input dimensions."""
        names = self.getInputNames()
        units = self.getInputUnits()
        scales = self.getInputScales()
        return ng.CoordinateSpace({
            name: [scale, unit]
            for name, scale, unit in zip(names, scales, units)
        })

    def getOutputDimensions(self) -> ng.CoordinateSpace:
        """Compute output dimensions."""
        names = self.getOutputNames()
        units = self.getOutputUnits()
        scales = self.getOutputScales()
        return ng.CoordinateSpace({
            name: [scale, unit]
            for name, scale, unit in zip(names, scales, units)
        })

    def getMatrix(self) -> np.ndarray:
        """Compute transformation matrix."""
        rank = self.getRank()
        return np.eye(rank+1)[:-1]

    def getTransform(self) -> ng.CoordinateSpaceTransform:
        """Compute neuroglancer transform."""
        return ng.CoordinateSpaceTransform(
            matrix=self.getMatrix(),
            input_dimensions=self.getInputDimensions(),
            output_dimensions=self.getOutputDimensions(),
        )


class VolumeDataSource(LayerDataSource):
    """Wrapper for volume source."""

    info: VolumeInfo

    @property
    def _transform(self) -> list[int]:
        return self.info.getTransform()

    @property
    def _nb_levels(self) -> list[int]:
        return self.info.getNbLevels()

    @property
    def _shape(self) -> list[int]:
        return self.info.getShape()

    @property
    def _dtype(self) -> np.dtype:
        return self.info.getDataType()

    @property
    def _slope_inter(self) -> tuple[float, float]:
        return self.info.getSlopeInter()


class SkeletonDataSource(LayerDataSource):
    """Wrapper for skeleton source."""

    ...


class MeshDataSource(LayerDataSource):
    """Wrapper for mesh source."""

    ...


class AnnotationDataSource(LayerDataSource):
    """Wrapper for annotation source."""

    ...


class LocalDataSource(LayerDataSource):
    """Wrapper for local data sources."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._format_specific_init()

    def _format_specific_init(self) -> None:
        pass


class LocalAnnotationDataSource(AnnotationDataSource, LocalDataSource):
    """Wrapper for local annotation sources."""

    ...


class _LocalSkeletonDataSourceFactory(_LayerDataSourceFactory):

    def __call__(
        cls,
        arg: _LayerDataSourceFactory._DataSourceLike = None,
        *args,
        **kwargs
    ) -> "LocalSkeletonDataSource":
        # Only use the factory if it is not called from a subclass
        if cls is not LocalSkeletonDataSource:
            obj = super().__call__(arg, *args, **kwargs)
            return obj

        # Get url
        if kwargs.get("url", ""):
            url = kwargs["url"]
            if isinstance(url, (str, PathLike)):
                url = kwargs["url"] = str(url)
        elif isinstance(arg, (str, PathLike)):
            url = arg = str(arg)
        elif hasattr(arg, "url"):
            url = arg.url
        elif isinstance(arg, dict) and "url" in arg:
            url = arg["url"]
        elif not isinstance(arg, cls._LocalSource):
            raise ValueError("Missing data source url")

        # If local object -> delegate
        if isinstance(url, ng.LocalVolume):
            raise ValueError(
                "Non skeleton local object passed to a skeleton source"
            )
        if isinstance(url, ng.skeleton.SkeletonSource):
            return super().__call__(arg, *args, **kwargs)

        parsed_url = parse_protocols(url)

        # If python:// url -> local object, but retrieved from the viewer.
        if parsed_url.stream == "python":

            # Get local object
            path = parsed_url.url
            source_type, token = path.split("://")[-1].strip("/").split("/")
            if source_type != "skeleton":
                raise ValueError(
                    "Non skeleton local object passed to a skeleton source"
                )
            vol = ng.server.global_server.get_volume(token)

            # Replace url with object
            if kwargs.get("url", ""):
                kwargs["url", vol]
            elif isinstance(arg, (str, PathLike)):
                arg = vol
            elif hasattr(arg, "url"):
                arg.url = vol
            elif isinstance(arg, dict) and "url" in arg:
                arg["url"] = vol

            # Return object
            kls = getattr(vol, "DataSourceType", LocalSkeletonDataSource)
            if kls is LocalSkeletonDataSource:
                return super().__call__(arg, *args, **kwargs)
            else:
                return kls(arg, *args, **kwargs)

        # If format protocol provided, use it
        if parsed_url.format in _DATASOURCE_REGISTRY:
            format = parsed_url.format
            LOG.debug(f"LocalSkeletonDataSource - use format hint: {format}")
            return _DATASOURCE_REGISTRY[format](arg, *args, **kwargs)

        # Otherwise, check for extensions
        for format, kls in _DATASOURCE_REGISTRY.items():
            if parsed_url.url.endswith((format, format+".gz", format+".bz2")):
                LOG.debug(f"LocalSkeletonDataSource - extension: {format}")
                try:
                    obj = kls(arg, *args, **kwargs)
                    LOG.debug(f"LocalSkeletonDataSource - {format} (success)")
                    return obj
                except Exception:
                    continue

        # Otherwise. build a simple source
        LOG.debug("LocalSkeletonDataSource - Fallback")
        return super().__call__(arg, *args, **kwargs)


class LocalSkeletonDataSource(SkeletonDataSource, LocalDataSource,
                              metaclass=_LocalSkeletonDataSourceFactory):
    """Wrapper for data source that wraps a `SkeletonSource`."""

    @property
    def local_skeleton(self) -> ng.LocalVolume | None:
        """Points to the underlying `SkeletonSource`, if any."""
        url = super().url
        if isinstance(self, ng.skeleton.SkeletonSource):
            return url
        return None

    ...


class _LocalVolumeDataSourceFactory(_LayerDataSourceFactory):

    def __call__(
        cls,
        arg: _LayerDataSourceFactory._DataSourceLike = None,
        *args,
        **kwargs
    ) -> "LocalVolumeDataSource":
        # Only use the factory if it is not called from a subclass
        if cls is not LocalVolumeDataSource:
            obj = super().__call__(arg, *args, **kwargs)
            return obj

        # Get url
        if kwargs.get("url", ""):
            url = kwargs["url"]
            if isinstance(url, (str, PathLike)):
                url = kwargs["url"] = str(url)
        elif isinstance(arg, (str, PathLike)):
            url = arg = str(arg)
        elif hasattr(arg, "url"):
            url = arg.url
        elif isinstance(arg, dict) and "url" in arg:
            url = arg["url"]
        elif not isinstance(arg, cls._LocalSource):
            raise ValueError("Missing data source url")

        # If local object -> delegate
        if isinstance(url, ng.LocalVolume):
            return LocalVolumeDataSource(arg, *args, **kwargs)
        if isinstance(url, ng.skeleton.SkeletonSource):
            raise ValueError(
                "Non volume local object passed to a volume source"
            )

        parsed_url = parse_protocols(url)

        # If python:// url -> local object, but retrieved from the viewer.
        if parsed_url.stream == "python":

            # Get local object
            path = parsed_url.url
            source_type, token = path.split("://")[-1].strip("/").split("/")
            if source_type != "volume":
                raise ValueError(
                    "Non volume local object passed to a volume source"
                )
            vol = ng.server.global_server.get_volume(token)

            # Replace url with object
            if kwargs.get("url", ""):
                kwargs["url", vol]
            elif isinstance(arg, (str, PathLike)):
                arg = vol
            elif hasattr(arg, "url"):
                arg.url = vol
            elif isinstance(arg, dict) and "url" in arg:
                arg["url"] = vol

            # Return object
            kls = getattr(vol, "DataSourceType", LocalVolumeDataSource)
            if kls is LocalVolumeDataSource:
                return super().__call__(arg, *args, **kwargs)
            else:
                return kls(arg, *args, **kwargs)

        # If format protocol provided, use it
        if parsed_url.format in _DATASOURCE_REGISTRY:
            format = parsed_url.format
            LOG.debug(f"LocalVolumeDataSource - use format hint: {format}")
            return _DATASOURCE_REGISTRY[format](arg, *args, **kwargs)

        # Otherwise, check for extensions
        for format, kls in _DATASOURCE_REGISTRY.items():
            if parsed_url.url.endswith((format, format+".gz", format+".bz2")):
                LOG.debug(f"LocalVolumeDataSource - extension: {format}")
                try:
                    obj = kls(arg, *args, **kwargs)
                    LOG.debug(f"LocalVolumeDataSource - {format} (success)")
                    return obj
                except Exception:
                    continue

        # Otherwise. build a simple source
        LOG.debug("LocalVolumeDataSource - Fallback")
        return super().__call__(arg, *args, **kwargs)


class LocalVolumeDataSource(VolumeDataSource, LocalDataSource,
                            metaclass=_LocalVolumeDataSourceFactory):
    """Wrapper for data source that wraps a `LocalVolume`."""

    def __init__(self, *args, **kwargs) -> None:
        self._layer_type = kwargs.pop("layer_type", None)
        super().__init__(*args, **kwargs)

    @property
    def local_volume(self) -> ng.LocalVolume | None:
        """Points to the underlying `LocalVolume`, if any."""
        url = super().url
        if isinstance(self, ng.LocalVolume):
            return url
        return None

    @property
    def _nb_levels(self) -> int:
        return 1

    @property
    def _dtype(self) -> np.dtype:
        return self.local_volume.data.data.dtype

    @property
    def _shape(self) -> np.dtype:
        return self.local_volume.data.shape

    def _get_dataobj(self, level: int = 0, mode: str = "r") -> ArrayLike:
        return self.local_volume.data.data


class NiftiVolumeInfo(VolumeInfo):
    """Metadata of a NIfTI volume."""

    def __init__(
        self,
        url: str | PathLike,
        align_corner: bool = False,
        affine: str = "best",
    ) -> None:
        """
        Parameters
        ----------
        url : str
            URL to the file
        align_corner : bool
            If True, use neuroglancer's native behavior when computing
            the transform, which is to asume that (0, 0, 0) points to
            the corner of the first voxel. If False, use NIfTI's spec,
            which is to assume the (0, 0, 0) points to the center of the
            first voxel.
        affine : {"qform", "sform", "best", "base"}
            Which orientation matrix to use.
            Neuroglancer uses the qform by default.
        """
        super().__init__(url)
        self._align_corner = align_corner
        self._affine = affine
        self._nib_header: nib.nifti1.Nifti1Header = self._load()

    def _load(self) -> nib.nifti1.Nifti1Header | nib.nifti2.Nifti2Header:
        NiftiHeaders = (nib.nifti1.Nifti1Header, nib.nifti2.Nifti2Header)
        tic = time.time()
        url = parse_protocols(self.url).url
        for compression in ('infer', 'gzip', None):
            fileobj = open(url, compression=compression)
            for hdr_klass in NiftiHeaders:
                try:
                    try:
                        hdr = hdr_klass.from_fileobj(fileobj, check=True)
                    except HeaderDataError:
                        # Check if failure is due to "old" nifti-zarr magic
                        hdr = hdr_klass.from_fileobj(fileobj, check=False)
                        if hdr["magic"].item().decode() not in ("nz1", "nz2"):
                            raise
                    toc = time.time()
                    LOG.info(f"Load nii header: {toc-tic} s")
                    return hdr
                except Exception:
                    pass
        toc = time.time()
        LOG.info(f"Failed to load nii header: {toc-tic} s")
        raise RuntimeError("Failed to load file.")

    _allSourceNames = ["i", "j", "k", "m", "c^", "c1^", "c2^"]
    _allViewNames = ["x", "y", "z", "t", "c^", "c1^", "c2^"]

    def getShape(self) -> list[int]:
        """Compute array shape."""
        return list(self._nib_header.get_data_shape())

    def getInputNames(self) -> list[str]:
        """Compute names of input axes."""
        return self._allSourceNames[:self.getRank()]

    def getOutputNames(self) -> list[str]:
        """Compute names of output axes."""
        return self._allViewNames[:self.getRank()]

    def getSpaceUnit(self) -> str:
        """Compute spatial unit."""
        if not getattr(self, "_space_unit", None):
            unit = self._nib_header.get_xyzt_units()[0]
            if unit == "unknown":
                unit = "mm"
            unit = U.as_neuroglancer_unit(unit)
            self._space_unit = unit
        return self._space_unit

    def getTimeUnit(self) -> str:
        """Compute time unit."""
        if not getattr(self, "_time_unit", None):
            unit = self._nib_header.get_xyzt_units()[1]
            if unit == "unknown":
                unit = "sec"
            unit = U.as_neuroglancer_unit(unit)
            self._time_unit = unit
        return self._time_unit

    def getUnits(self) -> list[str]:
        """Compute all units."""
        all_units = (
            [self.getSpaceUnit()] * 3 +
            [self.getTimeUnit()] +
            [""] * 3
        )
        return all_units[:self.getRank()]

    getInputUnits = getUnits
    getOutputUnits = getUnits

    def getInputScales(self) -> list[float]:
        """Compute input scales."""
        return list(map(float, self._nib_header.get_zooms()))

    def getOutputScales(self) -> list[float]:
        """Compute output scales."""
        return [1] * self.getRank()

    def getDescription(self) -> str:
        """Compute description string."""
        return self._nib_header["description"]

    def getDataType(self) -> np.dtype:
        """Compute numpy data type."""
        return self._nib_header.get_data_dtype()

    def getSlopeInter(self) -> tuple[float, float]:
        """Compute intensity transform."""
        scale, slope = self._nib_header.get_slope_inter()
        if scale is None:
            scale = 1.0
        if slope is None:
            slope = 0.0
        return scale, slope

    def getMatrix(self) -> np.ndarray:
        """Compute transformation matrix."""
        # Neuroglancer uses the qform (rotation), but without
        # scaling it by voxel size. The linear part therefore maps
        # spatial units to spatial units (m, mm or um)
        # and the offset is expressed in spatial units.
        #
        # Nibabel returns a voxel-to-spatial-unit transform, so we need
        # to rescale it
        #
        # If `_align_corner=False`, we introduce an additional shift
        # to account for the fact that neuroglancer assumes that coordinate
        # (0, 0, 0) of the input space is at the corner of the first voxel.
        if getattr(self, "_matrix", None) is None:
            srank = min(3, self.getRank())
            for affine in (self._affine, "best", "sform", "qform", "base"):
                if affine == "base":
                    matrix = self._nib_header.get_base_affine()[:-1]
                elif affine == "best":
                    matrix = self._nib_header.get_best_affine()[:-1]
                elif affine == "sform":
                    matrix = self._nib_header.get_sform()[:-1]
                else:
                    matrix = self._nib_header.get_qform()[:-1]
                scales = (matrix[:3, :3] ** 2).sum(0) ** 0.5
                if scales.max() == 0:
                    continue
                else:
                    break
            if not self._align_corner:
                matrix[:3, -1] -= matrix[:3, :-1] @ ([0.5]*srank)
            matrix[:3, :-1] /= scales
            fullmatrix = np.eye(self.getRank()+1)[:-1]
            fullmatrix[:srank, :srank] = matrix[:srank, :srank]
            fullmatrix[:srank, -1] = matrix[:srank, -1]
            self._matrix = fullmatrix
        return self._matrix


@datasource(["nifti", "nii"])
class NiftiDataSource(VolumeDataSource):
    """
    Wrapper for nifti sources.

    Note that by default, neuroglancer assumes that the coordinate
    (0, 0, 0) in voxel space corresponds to the _corner_ of the first
    voxel, whereas the nifti spec is that it corresponds to the _center_
    of the first voxel. This wrapper follows the nifti convention by default.
    The neuroglancer default behavior can be recovered with the option
    `align_corner=True`.

    Also, by default, neuroglancer uses the `qform` orientation matrix,
    whereas the `intent` code must be used to choose between the `sform`
    and `qform` according to the spec (as implemented in e.g. nibabel).
    This wrapper follows the specification, unless the affine to use
    is explictly specified with the option `affine="sform"`,
    `affine="qform"`, or `affine="base"`.
    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Parameters
        ----------
        json_data : dict | NiftiDataSource | str, optional
            URL to the file

        Other Parameters
        ----------------
        url : str
            URL to the file
        align_corner : bool, default=False
            If True, use neuroglancer's native behavior when computing
            the transform, which is to asume that (0, 0, 0) points to
            the corner of the first voxel. If False, use NIfTI's spec,
            which is to assume the (0, 0, 0) points to the center of the
            first voxel.
        affine : {"qform", "sform", "best", "base"}
            Which orientation matrix to use.
            Neuroglancer uses the qform by default.
        """
        self._align_corner = kwargs.pop("align_corner", False)
        self._select_affine = kwargs.pop("affine", "best")
        self._nib_image = None
        self._stream = None
        super().__init__(*args, **kwargs)
        # Trigger transform computation because we do not trust
        # neuroglancer's behaviour completely.
        self.transform = self.transform

    def _compute_info(self) -> NiftiVolumeInfo:
        return NiftiVolumeInfo(
            self.url,
            align_corner=self._align_corner,
            affine=self._select_affine,
        )

    def _load_image(self, mode: str = "r") -> nib.nifti1.Nifti1Image:
        tic = time.time()
        NiftiImages = (nib.nifti1.Nifti1Image, nib.nifti2.Nifti2Image)

        url = parse_protocols(self.url).url
        for compression in ('infer', 'gzip', None):
            self._stream = open(url, compression=compression)
            for img_klass in NiftiImages:
                try:
                    hdr = img_klass.from_stream(self._stream)
                    toc = time.time()
                    LOG.info(f"Loaded nifti file: {toc-tic} s")
                    return hdr
                except Exception:
                    pass
            self._stream = None
        toc = time.time()
        LOG.error(f"Failed to load nifti file {url}: {toc-tic} s")
        raise RuntimeError(f"Failed to load nifti file {url}.")

    def __del__(self) -> None:
        try:
            self._stream.close()
        except Exception:
            pass

    def _get_dataobj(self, *a, **k) -> ArrayLike:
        # NOTE: we return the raw (unscaled) data, like neuroglancer
        if getattr(self, "_nib_image", None) is None:
            self._nib_image = self._load_image()
        return self._nib_image.dataobj


class ZarrVolumeInfo(VolumeInfo):
    """Base class + common methods for Zarr metadata."""

    nifti: NiftiVolumeInfo | None

    def getShape(self, level: int = 0) -> list[int]:
        """Array shape at a given level."""
        return self._zarray[level]["shape"]

    def getNbLevels(self) -> int:
        """Return the number of levels in the pyramid."""
        try:
            return len(self.getMultiscalesOME()[0]["datasets"])
        except Exception:
            return 1

    def hasOME(self) -> bool:
        """Return `True` if Zarr has OME metadata."""
        return "multiscales" in self._zattrs or "ome" in self._zattrs

    def hasNIfTI(self) -> bool:
        """Return `True` if Zarr has NIfTI metadata."""
        return self.nifti is not None

    def getMultiscalesOME(self) -> list[dict]:
        """Return content of OME metadata (JSON)."""
        if "multiscales" in self._zattrs:
            return self._zattrs["multiscales"]
        else:
            return self._zattrs["ome"]["multiscales"]

    def getAxesOME(self) -> list[dict] | None:
        """Return OME axes (JSON)."""
        try:
            return self.getMultiscalesOME()[0]["axes"]
        except Exception:
            return None

    def getDatasetOME(self, level: int = 0) -> dict | None:
        """Return OME dataset (JSON)."""
        try:
            ome = self.getMultiscalesOME()[0]
            return ome["datasets"][level]
        except Exception:
            return None

    def getNames(self) -> list[str]:
        """Return axis names."""
        axes = self.getAxesOME()
        if axes is None:
            if isinstance(self, Zarr2VolumeInfo):
                return [f"d{i}" for i in range(self.getRank())]
            else:
                assert isinstance(self, Zarr3VolumeInfo)
                return [f"dim{i}" for i in range(self.getRank())]
        return [
            axis["name"] + ("'" if axis["type"] == "channel" else "")
            for axis in axes
        ]

    getInputNames = getOutputNames = getNames

    def getAxisTypes(self) -> list[str]:
        """Return axis type (space, time, channel)."""
        axes = self.getAxesOME()
        if axes is None:
            return [""] * self.getRank()
        return [axis["type"] for axis in axes]

    def getUnits(self) -> list[str]:
        """Return units."""
        axes = self.getAxesOME()
        if axes is None:
            return [""] * self.getRank()
        return [
            U.as_neuroglancer_unit(axis.get("unit", ""))
            for axis in axes
        ]

    getInputUnits = getOutputUnits = getUnits

    def getScales(self) -> list[float]:
        """Return scales."""
        if getattr(self, "_scales", None) is not None:
            return list(self._scales)

        scale = [1.0] * self.getRank()
        # Read common scale (applies to time)
        try:
            for trf in self.getMultiscalesOME()["coordinateTransformations"]:
                if "scale" in trf:
                    scale = trf["scale"]
                    break
        except Exception:
            pass
        # Read level 0 scale (applies to space)
        try:
            level0 = self.getDatasetOME()
            for transform in level0.get("coordinateTransformations", []):
                if "scale" in transform:
                    scale = [
                        scale1 if scale1 != 1 else scale0
                        for scale0, scale1 in zip(scale, transform["scale"])
                    ]
                    break
        except Exception:
            pass
        self._scales = scale
        return list(scale)

    getInputScales = getOutputScales = getScales

    def getTranslations(self) -> list[float]:
        """
        Return translations (in spatial units).

        Although OME-NGFF v0.4 requires translations to appear after
        scales so that translations are expressed in world units,
        neuroglancer accepts transformations that have the opposite
        order, i.e., where translations are expressed in voxels. We also
        handle this case and rescale translations to world units.
        """
        if getattr(self, "_translations", None) is not None:
            return list(self._translations)

        trans = [0.0] * self.getRank()
        # Read common scale (applies to time)
        try:
            root = self.getMultiscalesOME()[0]
            for trf in root["coordinateTransformations"]:
                if "scale" in trf:
                    scale = trf["scale"]
                    trans = [t0*s for t0, s in zip(trans, scale)]
                elif "translation" in trf:
                    trans = [t0+t for t0, t in zip(trans, trf["translation"])]
        except Exception:
            pass
        # Read level 0 scale (applies to space)
        try:
            level0 = self.getDatasetOME()
            for trf in level0.get("coordinateTransformations", []):
                if "scale" in trf:
                    scale = trf["scale"]
                    trans = [t0*s for t0, s in zip(trans, scale)]
                elif "translation" in trf:
                    trans = [t0+t for t0, t in zip(trans, trf["translation"])]
        except Exception:
            pass
        # OME defines translations with respect to the center of the corner
        # voxel, but neuroglancer defines them with respect to its corner.
        # We must therefore introduce an extra shift. See:
        # https://github.com/google/neuroglancer/blob/master/src/datasource/zarr/ome.ts#L244
        if self.hasOME:
            scales = self.getScales()
            types = self.getAxisTypes()
            trans = [
                t - 0.5 * (s if type == "space" else 0)
                for t, s, type in zip(trans, scales, types)
            ]
        self._translations = trans
        return list(trans)

    def getMatrix(self) -> np.ndarray:
        """Return transformation matrix."""
        if getattr(self, "_matrix", None) is None:
            matrix = np.eye(self.getRank()+1)[:-1]
            matrix[:, -1] = self.getTranslations()
            self._matrix = matrix
        return self._matrix

    def getZarrTransform(self) -> ng.CoordinateSpaceTransform:
        """Return the pure zarr transform."""
        return ng.CoordinateSpaceTransform(
            matrix=self.getMatrix(),
            input_dimensions=self.getInputDimensions(),
            output_dimensions=self.getOutputDimensions(),
        )

    def getNiftiInputDimensions(self) -> ng.CoordinateSpace:
        """Return nifti dimensions, but using OME axis names."""
        names = list(reversed(self.getNames()))
        units = self.nifti.getUnits()
        scales = self.nifti.getInputScales()
        return ng.CoordinateSpace({
            name: [scale, unit]
            for name, scale, unit in zip(names, scales, units)
        })

    def getNiftiOutputDimensions(self) -> ng.CoordinateSpace:
        """Return nifti output dimensions."""
        return self.nifti.getOutputDimensions()

    def getNiftiMatrix(self) -> np.ndarray:
        """Return nifti transformation matrix."""
        return self.nifti.getMatrix()

    def getNiftiTransform(self) -> ng.CoordinateSpaceTransform:
        """Return nifti transform."""
        return ng.CoordinateSpaceTransform(
            matrix=self.getNiftiMatrix(),
            input_dimensions=self.getNiftiInputDimensions(),
            output_dimensions=self.getNiftiOutputDimensions(),
        )

    def getNiftiZarrTransform(self) -> ng.CoordinateSpaceTransform:
        """
        Return the transform to apply to a Zarr layer so that it's
        oriented according to the nifti metadata.
        """
        return T.compose(
            # zarr output voxel size
            ng.CoordinateSpaceTransform(
                input_dimensions=self.getOutputDimensions(),
                output_dimensions=self.getOutputDimensions(),
            ),
            # nifti transform, but mapping from (x,y,z,t,c) to (x,y,z,t,c)
            # instead of (i,j,k,m,c) to (x,y,z,t,c)
            ng.CoordinateSpaceTransform(
                matrix=self.getNiftiMatrix(),
                input_dimensions=self.getNiftiOutputDimensions(),
                output_dimensions=self.getNiftiOutputDimensions(),
            ),
            # zarr input voxel size
            ng.CoordinateSpaceTransform(
                input_dimensions=self.getInputDimensions(),
                output_dimensions=self.getInputDimensions(),
            ),
        )

    def getTransform(self) -> ng.CoordinateSpaceTransform:
        """
        Return the nifti-zarr transform if it exists, else the ome-zarr
        transform.
        """
        return (
            self.getNiftiZarrTransform() if self.hasNIfTI() else
            self.getZarrTransform()
        )


class Zarr2VolumeInfo(ZarrVolumeInfo):
    """Volume info for Zarr v2."""

    def __init__(self, url: str, nifti: bool | None = None) -> None:
        super().__init__(url)
        url = UPath(parse_protocols(self.url).url)

        if exists(url / ".zgroup"):
            is_group = True
        elif exists(url / ".zarray"):
            is_group = False
            self._zarray = [read_json(url / ".zarray")]
        else:
            raise ValueError("Not a zarr")
        if exists(url / ".zattrs"):
            self._zattrs = read_json(url / ".zattrs")
        else:
            self._zattrs = {}

        if is_group:
            self._zarray = []
            try:
                zattrs = self._zattrs
                if "ome" in self._zattrs:
                    zattrs = zattrs["ome"]
                datasets = zattrs["multiscales"][0]["datasets"]
                for dataset in datasets:
                    path = dataset["path"]
                    zpath = url / path / ".zarray"
                    if exists(zpath):
                        self._zarray.append(read_json(zpath))
                    else:
                        raise ValueError("Missing array in OME zarr:", path)
            except Exception:
                raise ValueError("Missing or invalid OME metadata.")

        self.nifti = None
        if nifti is not False:
            if exists(url / "nifti" / ".zarray"):
                url = url / "nifti" / "0"
                try:
                    self.nifti = NiftiVolumeInfo(
                        url, affine="best", align_corner=True
                    )
                except Exception:
                    if nifti is True:
                        raise
            elif nifti is True:
                raise FileNotFoundError("Cannot find nifti group in zarr.")

    def getDataType(self) -> np.dtype:
        """Array shape at a given level."""
        return np.dtype(self._zarray[0]["dtype"])


class Zarr3VolumeInfo(ZarrVolumeInfo):
    """Volume info for Zarr v3."""

    def __init__(self, url: str, nifti: bool | None = None) -> None:
        super().__init__(url)
        url = UPath(parse_protocols(self.url).url)

        if not exists(url / "zarr.json"):
            raise ValueError("Not a zarr")

        zbase = read_json(url / "zarr.json")
        self._zattrs = zbase.get("attributes", {})
        if zbase["node_type"] == "group":
            is_group = True
        else:
            assert zbase["node_type"] == "array"
            is_group = False
            self._zarray = zbase

        if is_group:
            try:
                self._zarray = []
                zattrs = self._zattrs
                if "ome" in self._zattrs:
                    zattrs = zattrs["ome"]
                datasets = zattrs["multiscales"][0]["datasets"]
                for dataset in datasets:
                    path = dataset["path"]
                    zpath = url / path / "zarr.json"
                    if exists(zpath):
                        self._zarray.append(read_json(zpath))
                    else:
                        raise ValueError("Missing array in OME zarr:", path)
            except Exception:
                raise ValueError("Missing or invalid OME metadata.")

        self.nifti = None
        self._load_nifti(url, nifti)

    def _load_nifti(self, url: UPath, nifti: bool | None):
        if nifti is False:
            return

        nifti_json = url / "nifti" / "zarr.json"

        if not nifti_json.exists():
            if nifti is True:
                raise FileNotFoundError("Cannot find nifti group in zarr.")
            return

        try:
            content = json.loads(nifti_json.read_text())
            sep = content["chunk_key_encoding"]["configuration"]["separator"]
        except (KeyError, json.JSONDecodeError):
            raise ValueError(
                "Failed to read dimension separator from NIfTI zarr metadata."
            )

        nifti_group = url / "nifti" / f"c{sep}0"
        if not nifti_group.exists():
            raise FileNotFoundError("Cannot find nifti group in zarr.")

        self.nifti = NiftiVolumeInfo(
            nifti_group,
            affine="best",
            align_corner=True
        )

    def getDataType(self) -> np.dtype:
        """Array shape at a given level."""
        return np.dtype(self._zarray[0]["data_type"])


class _ZarrDataSourceFactory(_LayerDataSourceFactory):
    _DataSourceLike = _LayerDataSourceFactory._DataSourceLike

    def __call__(
            cls, arg: _DataSourceLike = None, *args, **kwargs
    ) -> "ZarrDataSource":
        # Only use the factory if it is not called from a subclass
        if cls is not ZarrDataSource:
            obj = super().__call__(arg, *args, **kwargs)
            return obj

        # Get url
        if kwargs.get("url", ""):
            url = kwargs["url"]
            if isinstance(url, (str, PathLike)):
                url = kwargs["url"] = str(url)
        elif isinstance(arg, (str, PathLike)):
            url = arg = str(arg)
        elif hasattr(arg, "url"):
            url = arg.url
        elif isinstance(arg, dict) and "url" in arg:
            url = arg["url"]
        elif isinstance(arg, cls._LocalSource):
            raise TypeError(f"Cannot convert {type(arg)} to ZarrDataSource")
        else:
            raise ValueError("Missing data source url")

        format = parse_protocols(url).format
        LOG.debug(f"ZarrDataSource - hint: {format}")

        if format == "zarr2":
            LOG.debug("ZarrDataSource -> Zarr2DataSource")
            return Zarr2DataSource(arg, *args, **kwargs)

        if format == "zarr3":
            LOG.debug("ZarrDataSource -> Zarr3DataSource")
            return Zarr3DataSource(arg, *args, **kwargs)

        LOG.debug("ZarrDataSource - guess version...")
        version = cls.guess_version(url)
        LOG.debug(f"ZarrDataSource - guess version: {version}")

        if version == 2:
            LOG.debug("ZarrDataSource -> Zarr2DataSource")
            return Zarr2DataSource(arg, *args, **kwargs)

        elif version == 3:
            LOG.debug("ZarrDataSource -> Zarr3DataSource")
            return Zarr3DataSource(arg, *args, **kwargs)

        LOG.debug("ZarrDataSource - fallback to super")
        return super().__call__(arg, *args, **kwargs)


@datasource(["zarr", "zarr2", "zarr3"])
class ZarrDataSource(VolumeDataSource, metaclass=_ZarrDataSourceFactory):
    """Wrapper for Zarr data sources.

    This wrapper will automatically check for the presence of a `nifti`
    group in the zarr object, which indicates that the file really has
    the nifti-zarr format.

    This behavior can be changed by setting the option `nifti=False`
    (to load the file as a pure OME-Zarr) or `nifti=True` (which raises
    an error if the nifti group is not found).
    """

    info: ZarrVolumeInfo

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._nifti = kwargs.pop('nifti', None)
        self._align_corner = kwargs.pop("align_corner", False)
        # trigger computation of the transform, in case it's a nifti-zarr
        # TODO: check if nifti zarr first.
        self.transform = self.transform

    @classmethod
    def guess_version(cls, url: str) -> int:
        """Guess zarr version."""
        url = UPath(parse_protocols(url).url)
        if exists(url / ".zgroup"):
            return 2
        if exists(url / ".zarray"):
            return 2
        if exists(url / "zarr.json"):
            return 3
        return 0

    def _compute_info(self) -> Zarr2VolumeInfo:
        version = self.guess_version(self.url)
        if version == 2:
            return Zarr2VolumeInfo(self.url)
        elif version == 3:
            return Zarr3VolumeInfo(self.url)

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def _get_dataobj(self, level: int = 0, mode: str = "r") -> ArrayLike:
        """Return an array-like object pointing to a pyramid level."""
        if zarr is None:
            raise ImportError("zarr must be available to access zarr data")
        tic = time.time()
        url = parse_protocols(self.url).url
        fs = filesystem(url)
        store = zarr.storage.FSStore(url, fs=fs, mode=mode)
        if not self.info.hasOME():
            dataobj = zarr.open(store, mode)
        else:
            path = self.info.getDatasetOME(level)["path"]
            dataobj = zarr.open(store, mode, path=path)
        toc = time.time()
        LOG.info(f"Map zarr: {toc - tic} s")
        return dataobj


@datasource("zarr2")
class Zarr2DataSource(ZarrDataSource):
    """Wrapper for Zarr v2 data sources."""

    def _compute_info(self) -> Zarr2VolumeInfo:
        return Zarr2VolumeInfo(self.url)


@datasource("zarr3")
class Zarr3DataSource(ZarrDataSource):
    """Wrapper for Zarr v3 data sources."""

    def _compute_info(self) -> Zarr3VolumeInfo:
        return Zarr3VolumeInfo(self.url)


class N5VolumeInfo(VolumeInfo):
    """Base class + common methods for N5 metadata."""

    def __init__(self, url: str) -> None:
        super().__init__(url)
        url = UPath(parse_protocols(self.url).url)
        if exists(url / "attributes.json"):
            self._attributes = read_json(url / "attributes.json")
        else:
            self._attributes = {}

        if "dimensions" in self._attributes:
            self._is_multiscale = False
            self._arrays = [self._attributes]
            self._attributes = {}
        else:
            self._is_multiscale = True

        if self._is_multiscale:
            self._arrays = []
            for i in range(self.getNbLevels()):
                url_i = url / f"s{i}" / "attributes.json"
                if exists(url_i):
                    self._arrays.append(read_json(url_i))
                else:
                    self._arrays.append({})

    def getDataType(self) -> np.dtype:
        """Array data type."""
        try:
            return np.dtype(self._attributes["dataType"])
        except Exception:
            return np.dtype(self._arrays[0]["dataType"])

    def getShape(self, level: int = 0) -> list[int]:
        """Array shape at a given level."""
        return list(self._arrays[level]["dimensions"])

    def getNbLevels(self) -> int:
        """Return the number of levels in the pyramid."""
        try:
            return len(self._attributes["downsamplingFactors"])
        except Exception:
            return 1

    def getNames(self) -> list[str]:
        """Return axis names."""
        rank = self.getRank()
        try:
            names = list(self._attributes["axes"])
        except Exception:
            names = list(self._arrays[0]["axes"])
        if len(names) != rank:
            raise ValueError(f"Rank mismatch: {len(names)} != {rank}")
        return names

    getInputNames = getOutputNames = getNames

    def getUnits(self) -> list[str]:
        """Return units."""
        rank = self.getRank()
        units = ["nm"] * self.getRank()
        attrs = [self._attributes] + self._arrays[:1]
        for attr in attrs:
            if "units" in attr:
                units = list(attr["units"])
                break
            elif "pixelResolution" in attr:
                unit = attr["pixelResolution"].get("unit", "")
                unit = unit or "nm"
                units = [unit] * rank
                break
        if len(units) != rank:
            raise ValueError(f"Rank mismatch: {len(units)} != {rank}")
        return units

    getInputUnits = getOutputUnits = getUnits

    def getScales(self) -> list[float]:
        """Return scales."""
        rank = self.getRank()
        scales = [1.0] * rank
        attrs = [self._attributes] + self._arrays[:1]
        for attr in attrs:
            if "resolution" in attr:
                scales = list(attr["resolution"])
                break
            elif "pixelResolution" in attr:
                scales = list(attr["pixelResolution"]["dimensions"])
                break
        if len(scales) != rank:
            raise ValueError(f"Rank mismatch: {len(scales)} != {rank}")
        return scales

    getInputScales = getOutputScales = getScales

    def getTranslations(self) -> list[float]:
        """Return translations (in spatial units)."""
        return [0.0] * self.getRank()


@datasource(["n5"])
class N5DataSource(VolumeDataSource):
    """Wrapper for N5 data sources."""

    info: N5VolumeInfo

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def _compute_info(self) -> Zarr2VolumeInfo:
        return N5VolumeInfo(self.url)

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def _get_dataobj(self, level: int = 0, mode: str = "r") -> ArrayLike:
        """Return an array-like object pointing to a pyramid level."""
        tic = time.time()
        url = parse_protocols(self.url).url
        fs = filesystem(url)

        if zarr is None:
            raise ImportError("zarr v2 must be available to access N5 data")
        try:
            from zarr.n5 import N5FSStore
        except (ImportError, ModuleNotFoundError):
            if int(zarr.__version__.split(".")[0]) > 2:
                raise ImportError(
                    f"N5 is only available in zarr v2, but you have "
                    f"v{zarr.__version__} installed.")

        store = N5FSStore(url, fs=fs, mode=mode)
        path = str(level)
        if path[:1] != "s":
            path = "s" + path
        dataobj = dataobj = zarr.open(store, mode, path=path)
        toc = time.time()
        LOG.info(f"Map n5: {toc - tic} s")
        return dataobj


class _PrecomputedInfoFactory(type):
    def __call__(cls, url: str | dict) -> "PrecomputedInfo":
        info = cls._load_dict(url)
        return {
            "neuroglancer_multiscale_volume": PrecomputedVolumeInfo,
            "neuroglancer_multilod_draco": PrecomputedMeshInfo,
            "neuroglancer_legacy_mesh": PrecomputedLegacyMeshInfo,
            "neuroglancer_skeletons": PrecomputedSkeletonInfo,
            "neuroglancer_annotations_v1": PrecomputedAnnotationInfo,
        }[info["@type"]](info)


class PrecomputedInfo(metaclass=_PrecomputedInfoFactory):
    """Base class for metadate about precomputed objects."""

    def __init__(self, url: str | dict) -> None:
        self._info = self._load_dict(url)

    @classmethod
    def _load_dict(cls, url: str | dict) -> dict:
        if not isinstance(url, dict):
            if url.startswith("precomputed://"):
                url = url[14:]
            with open(url, "rb") as f:
                info = json.load(f)
        else:
            info = url
        return info


class PrecomputedMeshInfo(PrecomputedInfo):
    """Metadata of a precomputed mesh."""

    ...


class PrecomputedLegacyMeshInfo(PrecomputedInfo):
    """Metadata of a precomputed mesh (legacy)."""

    ...


class PrecomputedSkeletonInfo(PrecomputedInfo):
    """Metadata of a precomputed skeleton."""

    ...


class PrecomputedAnnotationInfo(PrecomputedInfo):
    """Metadata of a precomputed annotation."""

    ...


class PrecomputedVolumeInfo(PrecomputedInfo):
    """Metadata of a precomputed volume."""

    def getDataType(self) -> np.ndarray:
        """TODO."""
        return np.dtype(self._info["data_type"])

    def getShape(self, level: int = 0) -> list[int]:
        """TODO."""
        if level < 0:
            level = self.getNbLevels() + level
        if level < 0 or level >= self.getNbLevels():
            raise IndexError("Pyramid level not available:", level)

        shape = list(self._info["scales"][level]["size"])
        if self._info["num_channels"] > 1:
            shape += [self._info["num_channels"]]
        return shape

    def getRank(self) -> int:
        """TODO."""
        return 4 if self._info["num_channels"] > 1 else 3

    def getScales(self) -> list[float]:
        """TODO."""
        scales = list(self._info["scales"][0]["resolution"])
        scales += max(0, self.getRank()-3) * [1.0]
        return scales

    def getUnits(self) -> list[str]:
        """TODO."""
        return ["nm"] * 3 + max(0, self.getRank()-3) * [""]

    def getNames(self) -> list[str]:
        """TODO."""
        names = ["x", "y", "z"]
        if self.getRank() == 4:
            names += ["c"]
        return names

    def getDimensions(self) -> ng.CoordinateSpace:
        """TODO."""
        names = self.getNames()
        scales = self.getScales()
        units = self.getUnits()
        return ng.CoordinateSpace({
            name: [scale, unit]
            for name, scale, unit in zip(names, scales, units)
        })

    def getTransform(self) -> ng.CoordinateSpaceTransform:
        """TODO."""
        dimensions = self.getDimensions()
        return ng.CoordinateSpaceTransform(
            matrix=np.eye(self.getRank()+1)[:-1],
            input_dimensions=dimensions,
            output_dimensions=dimensions,
        )

    def getNbLevels(self) -> int:
        """TODO."""
        return len(self._info["scales"])

    def getVoxelOffset(self) -> list[int]:
        """TODO."""
        return list(self._info["voxel_offset"])

    def getChunkSize(self, level: int = 0, choice: int = 0) -> list[int]:
        """TODO."""
        return self._info["scales"][level]["chunk_sizes"][choice]


class _PrecomputedDataSourceFactory(_LayerDataSourceFactory):
    def __call__(cls, url: str | dict) -> "PrecomputedDataSource":
        info = cls._load_dict(url)
        return {
            "neuroglancer_multiscale_volume": PrecomputedVolumeDataSource,
            "neuroglancer_multilod_draco": PrecomputedMeshDataSource,
            "neuroglancer_legacy_mesh": PrecomputedLegacyMeshDataSource,
            "neuroglancer_skeletons": PrecomputedSkeletonDataSource,
            "neuroglancer_annotations_v1": PrecomputedAnnotationDataSource,
        }[info["@type"]](info)


@datasource("precomputed")
class PrecomputedDataSource(LayerDataSource,
                            metaclass=_PrecomputedDataSourceFactory):
    """Base wrapper for precomputed data sources."""

    ...


class PrecomputedMeshDataSource(MeshDataSource, PrecomputedDataSource):
    """Base wrapper for precomputed mesh sources."""

    ...


class PrecomputedLegacyMeshDataSource(MeshDataSource, PrecomputedDataSource):
    """Base wrapper for precomputed mesh (legacy) sources."""

    ...


class PrecomputedSkeletonDataSource(SkeletonDataSource, PrecomputedDataSource):
    """Base wrapper for precomputed skeleton sources."""

    ...


class PrecomputedAnnotationDataSource(
    AnnotationDataSource, PrecomputedDataSource
):
    """Base wrapper for precomputed annotations sources."""

    ...


class PrecomputedVolumeDataSource(VolumeDataSource, PrecomputedDataSource):
    """Wrapper for precomputed volumes."""

    def _compute_info(self) -> PrecomputedVolumeInfo:
        return PrecomputedVolumeInfo(self.url)

    def _get_dataobj(self, level: int = 0, mode: str = "r") -> ArrayLike:
        if CloudVolume is None:
            raise ImportError(
                "cloudvolume must be available to access precomputed data"
            )
        url = parse_protocols(self.url).url
        return CloudVolume(url, mip=level)



"""Specialized `SkeletonSource` that loads and serves tractography data."""

# stdlib
import random
from io import BytesIO
from os import PathLike
from types import GeneratorType as generator
from typing import Literal

# externals
import numpy as np
from nibabel.streamlines.tck import TckFile
from nibabel.streamlines.trk import TrkFile

# internals
from ngtools.datasources import LocalSkeletonDataSource, datasource
from ngtools.opener import open, parse_protocols, stringify_path

# optionals
try:
    import neuroglancer as ng
    import neuroglancer.skeleton as ngsk
except ImportError:
    import ngtools._nglite as ng
    import ngtools._nglite.skeleton as ngsk


@datasource(["trk", "tck", "tracts"])
class TractDataSource(LocalSkeletonDataSource):
    """Data source for local skeletons."""

    def __init__(self, *args, **kwargs) -> None:
        self.nb_tracts = kwargs.pop('nb_tracts', None)
        super().__init__(*args, **kwargs)

    def _format_specific_init(self) -> None:
        if not isinstance(self.url, ngsk.SkeletonSource):
            self._url = self.url
            self.url = TractSkeleton(self._url, max_tracts=self.nb_tracts)
            self.transform = self.transform

    @property
    def _transform(self) -> ng.CoordinateSpaceTransform:
        rank = len(self.url.dimensions.names)
        return ng.CoordinateSpaceTransform(
            input_dimensions=self.url.dimensions,
            output_dimensions=self.url.dimensions,
            matrix=np.eye(rank+1)[:-1],
        )


_ONCE = []


class TractSkeleton(ngsk.SkeletonSource):
    """
    Local tractography source loaded into a neuroglancer skeleton.

    This class reads a TRK stremalines file (using nibabel)
    and implements methods that allow serving and/or converting the
    streamlines data into neuroglancer's precomputed skeleton format.

    The skeleton format was originally implemented to render skeletonized
    volumetric segmentations, with a relatively small number of
    individual objects. Since tractography (especially high-resolution
    tractography) can generate millions of streamlines, saving each
    streamline as an individual skeleton is very ineficient -- both for
    querying and rendering.

    Instead, we combine all streamlines into a single skeleton. Still,
    the large number of streamlines slows down rendering quite a lot, so
    I currently sample `MAX_TRACTS` streamlines from the file to display
    (currently set to 10,000).

    There is no "multi-scale" representation of the tracts in neuroglancer
    (i.e., generate tracts with a smaller or larger number of edges
    based on the current rendering scale). Maybe this is something that
    we should ask be added to neuroglancer.

    If a segmentation of the tractogram is available, it would probably
    make sense to save the tracts belonging to each segment in a
    different skeleton. This would allow switching each segment on and
    off, and would allow segments to be displayed in different colors.

    I also save a unit-norm orientation vector for each vertex. Saving
    this information as `vertex_attribute` allows using it for rendering
    (e.g., it can be used to render orientation-coded tracts).

    The specification of the precomputed skeleton format is available here:
        https://github.com/google/neuroglancer/blob/master/
        src/neuroglancer/datasource/precomputed/skeletons.md
    """

    DataSourceType = TractDataSource

    DEFAULT_MAX_TRACTS = 1000

    def __init__(
        self,
        fileobj: str | PathLike | BytesIO,
        max_tracts: int = DEFAULT_MAX_TRACTS,
        format: Literal['tck', 'trk'] | None = None,
        **kwargs: dict
    ) -> None:
        """
        Parameters
        ----------
        fileobj : path or file-like
            TCK or TRK file
        max_tracts : int
            Maximum number of tracts to display
        format : [list of] {'tck', 'trk'}
            Format hint
        """
        fileobj = stringify_path(fileobj)
        if isinstance(fileobj, str):
            fileobj = parse_protocols(fileobj).url
        self.fileobj = fileobj
        self.max_tracts = max_tracts or self.DEFAULT_MAX_TRACTS

        format = format or ['trk', 'tck']
        if not isinstance(format, (list, tuple)):
            format = [format]
        format = list(format)
        if 'trk' not in format:
            format += ['trk']
        if 'tck' not in format:
            format += ['tck']
        self.format = format

        self.tractfile = None
        self.displayed_ids = None
        self.displayed_tracts = None
        self.displayed_orientations = None

        self._ensure_loaded(lazy=True)
        super().__init__(ng.CoordinateSpace(
            names=["x", "y", "z"],
            units="mm",
            scales=[1, 1, 1],
        ), **kwargs)
        self.vertex_attributes["orientation"] = ngsk.VertexAttributeInfo(
            data_type=np.float32,
            num_components=3,
        )

    # ------------------------------------------------------------------
    #   SkeletonSource - API
    # ------------------------------------------------------------------

    def __getitem__(self, id: int) -> np.ndarray:
        """Get a single tract."""
        self._ensure_loaded()
        return self.tractfile.streamlines[id]

    def __len__(self) -> int:
        """Total number of tracts."""
        self._ensure_loaded()
        return len(self.tractfile.streamlines)

    # ------------------------------------------------------------------
    #   SkeletonSource - implementation
    # ------------------------------------------------------------------

    def _ensure_loaded(self, lazy: bool = False) -> None:
        """Load tracts from file (if `lazy=True`, only load metadata)."""
        if self._is_loaded(lazy):
            return

        klasses = dict(tck=TckFile, trk=TrkFile)

        def load(f: BytesIO) -> None:
            for format in self.format:
                klass = klasses[format]
                errors = {}
                try:
                    self.tractfile = klass.load(f, lazy_load=lazy)
                    return
                except Exception as e:
                    errors[format] = e

            print('error', *errors.values())
            raise RuntimeError('\n'.join(
                [f'{format}: {errors[format]}' for format in self.format]))

        if isinstance(self.fileobj, (str, PathLike)):
            with open(self.fileobj) as f:
                load(f)
        else:
            load(f)

    def _is_loaded(self, lazy: bool = False) -> bool:
        """Check if file is loaded."""
        if not self.tractfile:
            # not loaded at all
            return False
        if isinstance(self.tractfile.streamlines, generator):
            # lazilty loaded
            return lazy
        return True

    def _filter(self) -> None:
        """Select `max_tracts` random tracts."""
        self._ensure_loaded()
        num_tracts = len(self.tractfile.streamlines)
        ids = list(range(num_tracts))
        random.seed(1234)
        random.shuffle(ids)
        self.displayed_ids = ids[:self.max_tracts]

    def _make_skeleton(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Merge all (filtered) streamlines into a single skeleton.

        Returns
        -------
        vertices : (N, 3) array
        edges : (M, 2) array
        orientation : (N, 3) array
        """
        if not self.displayed_ids:
            self._filter()

        num_vertices = num_edges = 0
        vertices = []
        edges = []
        orientations = []

        for id in self.displayed_ids:
            tract = self[id]
            orient = self.compute_orientation(tract)

            vertices.append(tract.astype('<f4'))
            edges.append(np.stack([
                np.arange(len(tract) - 1, dtype='<u4') + num_vertices,
                np.arange(1, len(tract), dtype='<u4') + num_vertices,
            ], -1))
            orientations.append(orient.astype('<f4'))
            num_vertices += len(tract)
            num_edges += len(tract) - 1

        vertices = np.concatenate(vertices)
        edges = np.concatenate(edges)
        orientations = np.concatenate(orientations)
        return vertices, edges, orientations

    @classmethod
    def compute_orientation(cls, tract: np.ndarray) -> np.ndarray:
        """Compute the orientation of a tract at each vertex."""
        # 1) compute directed orientation of each edge
        orient = tract[1:] - tract[:-1]
        # 2) compute directed orientation of each vertex as the
        #    length-weighted average of the orientations of its edges
        #    (we just add them, and normalize later)
        orient = np.concatenate([
            orient[:1],                 # first vertex:   only one edge
            orient[:-1] + orient[1:],   # other vertices: two edges
            orient[-1:],                # last vertex:    only one edge
        ], 0)
        # 3) make orientations unit-length
        length = np.sqrt((orient * orient).sum(-1, keepdims=True))
        length = np.clip(length, 1e-12, None)
        orient /= length
        return orient

    def get_skeleton(self, i: int = 1) -> ngsk.Skeleton:
        """
        Neuroglancer Python API.

        Parameters
        ----------
        i : int
            Segment index (should be zero in our case)

        Returns
        -------
        skel : neuroglancer.Skeleton
        """
        # if i != 1:
        #     raise ValueError('Unknown segment id')
        self._filter()
        vertices, edges, orients = self._make_skeleton()
        return ngsk.Skeleton(vertices, edges, dict(orientation=orients))

    # ------------------------------------------------------------------
    #   Converter to Neuroglancer precomputed skeleton format
    # ------------------------------------------------------------------

    def precomputed_prop_info(self, combined: bool = False) -> dict:
        """
        Return precomputed format's "property info" file.

            https://github.com/google/neuroglancer/blob/master/
            src/neuroglancer/datasource/precomputed/segment_properties.md
        """
        self._filter()
        num_tracts = len(self.displayed_ids)

        # if not combined, each tract correspond to a different "segment"
        # if combined, all tracts are merged into a single skeleton
        ids = ["1"] if combined else [str(i) for i in range(num_tracts)]

        info = {
            "@type": "neuroglancer_segment_properties",
            "inline": {
                "ids": ids,
                "properties": [{
                    "id": "label",
                    "type": "label",
                    "values": ids,
                }]
            }
        }
        return info

    def precomputed_skel_info(self) -> dict:
        """
        Return precomputed format's "skeleton info" file.

            https://github.com/google/neuroglancer/blob/master/
            src/neuroglancer/datasource/precomputed/skeletons.md
        """
        self._ensure_loaded(lazy=True)

        # No need for a transform, as nibabel automatically converts
        # TRK coordinates to mm RAS+
        # All we need to do is convert mm to nm (we do this in the
        # track serving functions)

        info = {
            "@type": "neuroglancer_skeletons",
            "vertex_attributes": [
                {
                    "id": "orientation",
                    "data_type": "float32",
                    "num_components": 3,
                },
            ],
            "segment_properties": "prop",
        }

        return info

    def precomputed_skel_data(
        self,
        id: int = 1,
        combined: bool = True,
    ) -> bytes:
        """Return precomputed format's "skeleton data" bytes."""
        if combined:
            return self._precomputed_skel_data_combined(id)
        else:
            return self._precomputed_skel_data_single(id)

    def _precomputed_skel_data_combined(self, id: int = 1) -> bytes:
        """
        Return precomputed format's "skeleton data" bytes (combined).

        `id` should alway be `1` (since there is only one skeleton)

        TODO: combine tracts per segment if a tractogram segmentation
        is available

            https://github.com/google/neuroglancer/blob/master/
            src/neuroglancer/datasource/precomputed/skeletons.md
            #encoded-skeleton-file-format
        """
        if id != 1:
            return b''
        self._filter()

        # skeleton format:
        # num_vertices:     uint32le
        # edges:            uint32le
        # vertex_positions: float32le [num_vertices, 3]  (C-order)
        # edges:            uint32le  [num_edges,    2]  (C-order)
        # attr|orientation: float32le [num_vertices, 3]  (C-order)

        v, e, o = self._make_skeleton()
        v *= 1E6

        bintract = b''
        bintract += np.asarray(len(v), dtype='<u4').tobytes()
        bintract += np.asarray(len(e), dtype='<u4').tobytes()
        bintract += np.asarray(v, dtype='<f4').tobytes()
        bintract += np.asarray(e, dtype='<u4').tobytes()
        bintract += np.asarray(o, dtype='<f4').tobytes()
        return bintract

    def _precomputed_skel_data_single(self, id: int) -> bytes:
        """
        Return precomputed format's "skeleton data" bytes (single tract).

        This function is used in the case where each streamline is
        encoded by a single skeleton. Note that this is very inefficient.

        TODO: add orientation attribute
        """
        self._filter()

        # skeleton format:
        # num_vertices:     uint32le
        # edges:            uint32le
        # vertex_positions: float32le [num_vertices, 3]  (C-order)
        # edges:            uint32le  [num_edges,    2]  (C-order)
        # attr|orientation: float32le [num_vertices, 3]  (C-order)

        tract = self.displayed_tracts[id] * 1E6
        v, e, o = self.get_skeleton(tract)

        bintract = b''
        bintract += np.asarray(len(v), dtype='<u4').tobytes()
        bintract += np.asarray(len(e), dtype='<u4').tobytes()
        bintract += np.asarray(v, dtype='<f4').tobytes()
        bintract += np.asarray(e, dtype='<u4').tobytes()
        bintract += np.asarray(o, dtype='<f4').tobytes()

        print('serve tract', id, '/', len(self.displayed_ids))
        return bintract



"""Read/write linear transforms."""
from pathlib import Path
import numpy as np
from nibabel import load as loadimg

from ..patched import LabeledWrapStruct


class TransformIOError(IOError):
    """General I/O exception while reading/writing transforms."""


class TransformFileError(TransformIOError):
    """Specific I/O exception when a file does not meet the expected format."""


class StringBasedStruct(LabeledWrapStruct):
    """File data structure from text files."""

    def __init__(self, binaryblock=None, endianness=None, check=True):
        """Create a data structure based off of a string."""
        _dtype = getattr(binaryblock, "dtype", None)
        if binaryblock is not None and _dtype == self.dtype:
            self._structarr = binaryblock.copy()
            return
        super().__init__(binaryblock, endianness, check)

    def __array__(self):
        """Return the internal structure array."""
        return self._structarr

    def to_string(self):
        """Convert to a string directly writeable to file."""
        raise NotImplementedError

    @classmethod
    def from_string(cls, string):
        """Read the struct from string."""
        raise NotImplementedError


class LinearTransformStruct(StringBasedStruct):
    """File data structure from linear transforms."""

    def to_filename(self, filename):
        """Store this transform to a file with the appropriate format."""
        with open(str(filename), "w") as f:
            f.write(self.to_string())

    def to_ras(self, moving=None, reference=None):
        """Return a nitransforms internal RAS+ matrix."""
        raise NotImplementedError

    @classmethod
    def from_filename(cls, filename):
        """Read the struct from a file given its path."""
        with open(str(filename)) as f:
            string = f.read()
        return cls.from_string(string)

    @classmethod
    def from_fileobj(cls, fileobj, check=True):
        """Read the struct from a file object."""
        return cls.from_string(fileobj.read())

    @classmethod
    def from_ras(cls, ras, moving=None, reference=None):
        """Create an affine from a nitransform's RAS+ matrix."""
        raise NotImplementedError


class LinearParameters(LinearTransformStruct):
    """
    A string-based structure for linear transforms.

    Examples
    --------
    >>> lp = LinearParameters()
    >>> np.array_equal(lp.structarr['parameters'], np.eye(4))
    True

    >>> p = np.diag([2., 2., 2., 1.])
    >>> lp = LinearParameters(p)
    >>> np.array_equal(lp.structarr['parameters'], p)
    True

    """

    template_dtype = np.dtype([("parameters", "f8", (4, 4))])
    dtype = template_dtype

    def __init__(self, parameters=None):
        """
        Initialize with default parameters.


        """
        super().__init__()
        self.structarr["parameters"] = np.eye(4)
        if parameters is not None:
            self.structarr["parameters"] = parameters


class BaseLinearTransformList(LinearTransformStruct):
    """A string-based structure for series of linear transforms."""

    template_dtype = np.dtype([("nxforms", "i4")])
    dtype = template_dtype
    _xforms = None
    _inner_type = LinearParameters

    def __init__(self, xforms=None, binaryblock=None, endianness=None, check=True):
        """Initialize with (optionally) a list of transforms."""
        super().__init__(binaryblock, endianness, check)
        self.xforms = [self._inner_type(parameters=mat) for mat in xforms or []]

    @property
    def xforms(self):
        """Get the list of internal transforms."""
        return self._xforms

    @xforms.setter
    def xforms(self, value):
        self._xforms = list(value)

    def __getitem__(self, idx):
        """Allow dictionary access to the transforms."""
        if idx == "xforms":
            return self._xforms
        if idx == "nxforms":
            return len(self._xforms)
        raise KeyError(idx)


class DisplacementsField:
    """A data structure representing displacements fields."""

    @classmethod
    def from_filename(cls, filename):
        """Import a displacements field from a NIfTI file."""
        imgobj = loadimg(str(filename))
        return cls.from_image(imgobj)

    @classmethod
    def from_image(cls, imgobj):
        """Import a displacements field from a nibabel image object."""
        raise NotImplementedError

    @classmethod
    def to_filename(cls, img, filename):
        """Export a displacements field to a NIfTI file."""
        imgobj = cls.to_image(img)
        imgobj.to_filename(filename)

    @classmethod
    def to_image(cls, imgobj):
        """Export a displacements field image from a nitransforms image object."""
        raise NotImplementedError


def _ensure_image(img):
    if isinstance(img, (str, Path)):
        return loadimg(img)
    return img



"""General-purpose byte stream opener.

Classes
-------
open
    General-purpose stream opener.
chain
    Chain multiple file objects.

Functions
---------
stringify_path
    Ensure that the input is a str or a file-like object.
parse_protocols
    Parse protocols out of a url.
fsopen
    Open a file with fsspec (authentify if needed).
filesystem
    Return the fsspec filesystem corresponding to a protocol or URI.
exists
    Check that the file or directory pointed by a URI exists.
linc_auth_opt
    Create fsspec authentication options to access lincbrain data.
read_json
    Read local or remote JSON file.
write_json
    Write local or remote JSON file.
update_json
    Update local or remote JSON file.
"""
# stdlib
import json
import logging
import time
from bz2 import BZ2File
from collections import namedtuple
from inspect import isawaitable
from io import BufferedReader, BytesIO
from os import PathLike, environ
from pathlib import Path
from types import TracebackType
from typing import IO

# externals
import fsspec
import fsspec.asyn
import numpy as np
import requests
from typing_extensions import Buffer

# internals
from ngtools.protocols import FORMATS, LAYERS, PROTOCOLS

# optionals
try:
    from indexed_gzip import IndexedGzipFile as GzipFile
except ImportError:
    from gzip import GzipFile

LOG = logging.getLogger(__name__)

DANDI_API = {
    "linc": "https://api.lincbrain.org/api",
    "dandi": "https://api.dandiarchive.org/api",
}

# Cache for LINC cookies
_LINC_AUTH_CACHE = {}
_DANDI_AUTH_CACHE = {}

# Cache for fsspec filesystems
_FILESYSTEMS_CACHE = {}

# Hints
URILike = PathLike | str
FileLike = IO | URILike
FileSystem = fsspec.AbstractFileSystem


def stringify_path(filename: FileLike) -> IO | str:
    """Ensure that the input is a str or a file-like object."""
    if isinstance(filename, PathLike):
        return filename.__fspath__()
    if isinstance(filename, Path):
        return str(filename)
    return filename


parsed_protocols = namedtuple(
    "parsed_protocols",
    ["layer", "format", "stream", "url"]
)


class parse_protocols(parsed_protocols):
    """Parse ngtools uri."""

    def __new__(cls, *args, **kwargs) -> "parse_protocols":  # noqa: D102
        layer = format = stream = url = None

        args = list(args)
        if args:
            url = str(args.pop(-1))
            *parts, url = str(url).split("://")
            for part in parts:

                if part in PROTOCOLS:
                    if stream is not None:
                        raise ValueError("Too many streaming protocols:",
                                         stream, part)
                    stream = part

                elif part in LAYERS:
                    if layer is not None:
                        raise ValueError("Too many layer protocols:",
                                         layer, part)
                    layer = part

                elif part in FORMATS:
                    if format is not None:
                        raise ValueError("Too many format protocols:",
                                         format, part)
                    format = part

                else:
                    raise ValueError("Unknown protocol:", part)

        for i, arg in enumerate(reversed(args)):
            if i == 0:
                stream = arg
            elif i == 1:
                format = arg
            elif i == 2:
                layer = arg
            else:
                raise ValueError("Too many inputs")
        layer = kwargs.get("layer", layer)
        format = kwargs.get("format", format)
        stream = kwargs.get("stream", stream)
        url = kwargs.get("url", url)

        stream = stream or "file"
        if stream != "file":
            url = stream + "://" + url
        return super().__new__(cls, layer, format, stream, url)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

    @property
    def path(self) -> str:
        """Alias for url."""
        return self.url

    def __str__(self) -> str:
        out = self.url
        if self.stream and not self.url.startswith(self.stream):
            out = self.stream + "://" + out
        if self.format:
            out = self.format + "://" + out
        if self.layer:
            out = self.layer + "://" + out
        return out

    def with_part(self, **kwargs) -> "parse_protocols":
        """Replace parts."""
        if "stream" not in kwargs:
            url = kwargs.get("url", self.url)
            if "://" in url:
                kwargs["stream"] = url.split("://")[0]
        return parse_protocols(
            kwargs.get("layer", self.layer),
            kwargs.get("format", self.format),
            kwargs.get("stream", self.stream),
            kwargs.get("url", self.url),
        )

    def with_layer(self, layer: str | None) -> "parse_protocols":
        """Replace layer."""
        return self.with_part(layer=layer)

    def with_format(self, format: str | None) -> "parse_protocols":
        """Replace format."""
        return self.with_part(format=format)

    def with_stream(self, stream: str | None) -> "parse_protocols":
        """Replace stream."""
        return self.with_part(stream=stream)

    def with_url(self, url: str | None) -> "parse_protocols":
        """Replace url."""
        return self.with_part(url=url)

    def with_path(self, path: str | None) -> "parse_protocols":
        """Replace path."""
        return self.with_part(url=path)


def linc_auth_opt(token: str | None = None) -> dict:
    """
    Create fsspec authentication options to access lincbrain data.

    These options should only be used when accessing data behind
    `neuroglancer.lincbrain.org`.

    Parameters
    ----------
    token : str
        Your LINCBRAIN_API_KEY

    Returns
    -------
    opt : dict
        options to pass to `fsspec`'s `HTTPFileSystem`.
    """
    API = DANDI_API['linc']
    token = token or environ.get('LINCBRAIN_API_KEY', None)
    if not token:
        return {}
    if token in _LINC_AUTH_CACHE:
        return _LINC_AUTH_CACHE[token]
    headers = {"Authorization": f"token {token}"}
    # Check that credential is correct
    session = requests.Session()
    session.get(f"{API}/auth/token", headers=headers)
    # Get cookies
    response = session.get(f"{API}/permissions/s3/", headers=headers)
    cookies = response.cookies.get_dict()
    # Pass cookies to FileSystem
    opt = {'cookies': cookies}
    _LINC_AUTH_CACHE[token] = opt
    return opt


def dandi_auth_opt(token: str | None = None, instance: str = "dandi") -> dict:
    """
    Create fsspec authentication options to access the dandi api.

    These options should only be used when accessing data behind
    `dandi://`.

    Parameters
    ----------
    token : str
        Your DANDI_API_KEY or LINCBRAIN_API_KEY
    instance : {"dandi", "linc"}

    Returns
    -------
    opt : dict
        options to pass to `fsspec`'s `HTTPFileSystem`.
    """
    prefix = {"dandi": "DANDI", "linc": "LINCBRAIN"}[instance]
    token = (
        token
        or environ.get(f'{prefix}_API_KEY', None)
        or environ.get('DANDI_API_KEY', None)
    )
    if not token:
        return {}
    if token in _DANDI_AUTH_CACHE:
        return _DANDI_AUTH_CACHE[token]
    headers = {"Authorization": f"token {token}"}
    # Check that credential is correct
    session = requests.Session()
    session.get(f"{DANDI_API[instance]}/auth/token", headers=headers)
    # Pass cookies to FileSystem
    opt = {"headers": headers}
    _DANDI_AUTH_CACHE[token] = opt
    return opt


def filesystem(protocol: URILike | FileSystem, **opt) -> FileSystem:
    """Return the filesystem corresponding to a protocol or URI."""
    if isinstance(protocol, fsspec.AbstractFileSystem):
        return protocol
    protocol = stringify_path(protocol)
    if protocol not in PROTOCOLS:
        protocol = parse_protocols(protocol)
        url = protocol.url
        protocol = protocol.stream

    # --- LINC/DANDI authentification ---
    linc_auth = opt.pop("linc_auth", None)
    if linc_auth is None:
        linc_auth = "neuroglancer.lincbrain.org" in url
        linc_auth = linc_auth or "dandi://linc" in url
    dandi_auth = opt.pop("dandi_auth", None)
    if dandi_auth is None:
        dandi_auth = "dandi://" in url
    # -----------------------------------

    if (protocol, linc_auth, dandi_auth) in _FILESYSTEMS_CACHE:
        return _FILESYSTEMS_CACHE[(protocol, linc_auth, dandi_auth)]

    # --- LINC/DANDI authentification ---
    if linc_auth:
        LOG.debug(f"linc_auth - {url}")
        opt.setdefault("client_kwargs", {})
        opt["client_kwargs"].update(linc_auth_opt())
    if dandi_auth:
        LOG.debug(f"dandi_auth - {url}")
        instance = "linc" if linc_auth else "dandi"
        opt.setdefault("client_kwargs", {})
        opt["client_kwargs"].update(dandi_auth_opt(instance=instance))
    # -----------------------------------

    fs = fsspec.filesystem(protocol, **opt)
    _FILESYSTEMS_CACHE[(protocol, linc_auth, dandi_auth)] = fs
    return fs


def exists(uri: URILike, **opt) -> bool:
    """Check that the file or directory pointed by a URI exists."""
    tic = time.time()
    uri0 = uri
    uri = parse_protocols(uri).url
    fs = filesystem(uri, **opt)
    exists = fs.exists(uri)
    toc = time.time()
    LOG.debug(f"exists({uri0}): {exists} | {toc-tic} s")
    return exists


async def async_exists(uri: URILike, **opt) -> bool:
    """Check that the file or directory pointed by a URI exists."""
    tic = time.time()
    uri0 = uri
    uri = parse_protocols(uri).url
    fs = filesystem(uri, **opt)
    if isinstance(fs, fsspec.asyn.AsyncFileSystem):
        exists = await fs._exists(uri)
    else:
        exists = fs.exists(uri)
    toc = time.time()
    LOG.debug(f"exists({uri0}): {exists} | {toc-tic} s")
    return exists


def read_json(fileobj: FileLike, *args, **kwargs) -> dict:
    """Read local or remote JSON file."""
    tic = time.time()
    with open(fileobj, "rb") as f:
        data = json.load(f, *args, **kwargs)
    toc = time.time()
    LOG.debug(f"read_json({fileobj}): {toc-tic} s")
    return data


def write_json(obj: dict, fileobj: FileLike, *args, **kwargs) -> None:
    """Write local or remote JSON file."""
    tic = time.time()
    with open(fileobj, "w") as f:
        json.dump(obj, f, *args, **kwargs)
    toc = time.time()
    LOG.debug(f"write_json(..., {fileobj}): {toc-tic} s")
    return


def update_json(obj: dict, fileobj: FileLike, *args, **kwargs) -> dict:
    """Read local or remote JSON file."""
    write_json(read_json(fileobj).update(obj), fileobj, *args, **kwargs)


def fsopen(
    uri: URILike,
    mode: str = "rb",
    compression: str | None = None
) -> fsspec.spec.AbstractBufferedFile:
    """Open a file with fsspec (authentify if needed)."""
    fs = filesystem(uri)
    return fs.open(uri, mode, compression=compression)


async def fsopen_async(
    uri: URILike,
    mode: str = "rb",
    compression: str | None = None
) -> fsspec.spec.AbstractBufferedFile:
    """Open a file with fsspec (authentify if needed)."""
    fs = filesystem(uri)
    if hasattr(fs, "open_async"):
        return await fs.open_async(uri, mode, compression=compression)
    else:
        return fs.open(uri, mode, compression=compression)


class chain:
    """Chain multiple file objects."""

    # FIXME: never tested and not really used in the end

    def __init__(self, *files, **kwargs) -> None:
        self._files = list(map(lambda x: open(files, **kwargs)))
        self._file_index = 0

    def _open(self) -> None:
        self._open(self._file_index)

    def _close(self) -> None:
        self._close(self._file_index)

    def __enter__(self) -> IO:
        self._open()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._close()

    def read(self, n: int | None = None, *a, **k) -> bytes | str:
        """Read characters ("t") or byes ("b")."""
        out = None
        while self._file_index < len(self._files):
            with self._files[self._file_index] as f:
                out1 = f.read(n, *a, **k)
                if n != 0 and len(out1) == 0:
                    self._file_index += 1
                if n is not None:
                    n -= len(out1)
                out = out1 if out is None else (out + out1)
                if n <= 0:
                    break
        return out

    def readline(self, n: int | None = None, *a, **k) -> bytes | str:
        """Read lines ("t")."""
        out = []
        while self._file_index < len(self._files):
            with self._files[self._file_index] as f:
                out1 = f.readlines(n, *a, **k)
                if n != 0 and len(out1) == 0:
                    self._file_index += 1
                if n is not None:
                    n -= len(out1)
                out += out1
                if n <= 0:
                    break
        return out

    def readinto(self, buf: Buffer, *a, **k) -> None:
        """Read bytes into an existing buffer."""
        # FIXME: I am not sure this implementation works
        buf0 = None
        if isinstance(buf, np.ndarray):
            buf0, buf = buf, np.empty_like(buf, shape=[buf.numel()])

        count = 0
        while self._file_index < len(self._files):
            with self._files[self._file_index] as f:
                count1 = f.readinto(buf[count:], *a, **k)
                if len(buf) != 0 and count1 in (0, None):
                    self._file_index += 1
                if count1 is not None:
                    count += count1
                if count >= len(buf):
                    break

        if buf0 is not None:
            buf0[...] = buf.reshape(buf0.shape)
            buf = buf0
        return count

    def write(self, *a, **k) -> int:
        """Write characters ("t") or bytes ("b")."""
        raise NotImplementedError

    def writeline(self, *a, **k) -> int:
        """Write lines."""
        raise NotImplementedError

    def peek(self, *a, **k) -> bytes | str:
        """Read tge first bytes without moving the cursor."""
        head = self.read(*a, **k)
        self._files.insert(self._file_index, BytesIO(head))
        return head

    def seek(self, *a, **k) -> int:
        """Move the cursor."""
        raise NotImplementedError

    def tell(self, *a, **k) -> int:
        """Return the cursor position."""
        raise NotImplementedError


class open:
    """
    General-purpose stream opener.

    Handles:

    * paths (local or url)
    * opened file-objects
    * anything that `fsspec` handles
    * compressed streams
    """

    def __init__(
        self,
        fileobj: FileLike,
        mode: str = 'rb',
        compression: str | None = None,
        open: bool = True,
    ) -> None:
        """
        Open a file from a path or url or an opened file object.

        Parameters
        ----------
        fileobj : path or url or file-object
            Input file
        mode : str
            Opening mode
        compression : {'zip', 'bz2', 'gzip', 'lzma', 'xz', 'infer'}
            Compression mode.
            If 'infer', guess from magic number (if mode 'r') or
            filename (if mode 'w').
        auth : dict
            Authentification options to pass to `fsspec`.

        Returns
        -------
        fileobj
            Opened file
        """
        self.fileobj: str | IO = stringify_path(fileobj)
        """The input path or file-like object"""
        self.fileobjs: list[IO] = []
        """Additional file-like objects that may be created upon opening."""
        self.mode: str = mode
        """Opening mode."""
        self.compression: str = compression
        """Compression mode."""

        # open
        self._is_open = False
        if open:
            self._open()

    @property
    def _effective_fileobj(self) -> IO:
        # This is the most final file-like object
        return self.fileobjs[-1] if self.fileobjs else self.fileobj

    def _close(self) -> None:
        # close all the file-like objects that we've created
        for fileobj in reversed(self.fileobjs):
            fileobj.close()
        self.fileobjs = []
        self._is_open = False

    def _open(self) -> None:
        if self._is_open:
            return
        if isinstance(self.fileobj, str):
            self._fsspec_open()
        if self.compression == 'infer' and 'r' in self.mode:
            self._infer()
        self._is_open = True

    def _fsspec_open(self) -> None:
        # Take the last file object (most likely the input fileobj), which
        # must be a path, and open it with `fsspec.open()`.
        uri = stringify_path(self._effective_fileobj)
        # Ensure that input is a string
        if not isinstance(uri, str):
            raise TypeError("Expected a URI, but got:", uri)
        # Set compression option
        opt = dict()
        if not (self.compression == 'infer' and 'r' in self.mode):
            opt['compression'] = self.compression
        # Open with fsspec
        fileobj = fsopen(uri, mode=self.mode, **opt)
        self.fileobjs.append(fileobj)
        # -> returns a fsspec object that's not always fully opened
        #    so open again to be sure.
        if hasattr(fileobj, 'open'):
            fileobj = fileobj.open()
            self.fileobjs.append(fileobj)

    def _infer(self) -> None:
        # Take the last file object (which at this point must be a byte
        # stream) and infer its compression from its magic bytes.
        fileobj = self._effective_fileobj
        if not hasattr(fileobj, "peek"):
            fileobj = BufferedReader(fileobj)
            self.fileobjs.append(fileobj)
        try:
            magic = fileobj.peek(2)
        except Exception:
            return
        if magic == b'\x1f\x8b':
            fileobj = GzipFile(fileobj)
            self.fileobjs.append(fileobj)
        if magic == b'BZh':
            fileobj = BZ2File(fileobj)
            self.fileobjs.append(fileobj)

    # ------------------------------------------------------------------
    # open() / IO API
    # ------------------------------------------------------------------

    def __enter__(self) -> IO:
        self._open()
        return self._effective_fileobj

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._close()

    def __del__(self) -> None:
        self._close()

    def open(self) -> IO:
        """Open the file."""
        if not self._is_open:
            self._open()
        return self

    def close(self) -> None:
        """Close the file object."""
        self._close()

    def read(self, *a, **k) -> bytes | str:
        """Read characters ("t") or byes ("b")."""
        return self._effective_fileobj.read(*a, **k)

    def readline(self, *a, **k) -> bytes | str:
        """Read lines ("t")."""
        return self._effective_fileobj.readline(*a, **k)

    def readinto(self, *a, **k) -> None:
        """Read bytes into an existing buffer."""
        return self._effective_fileobj.readinto(*a, **k)

    def write(self, *a, **k) -> int:
        """Write characters ("t") or bytes ("b")."""
        return self._effective_fileobj.write(*a, **k)

    def writeline(self, *a, **k) -> int:
        """Write lines."""
        return self._effective_fileobj.writeline(*a, **k)

    def peek(self, *a, **k) -> bytes | str:
        """Read the first bytes without moving the cursor."""
        return self._effective_fileobj.peek(*a, **k)

    def seek(self, *a, **k) -> int:
        """Move the cursor."""
        return self._effective_fileobj.seek(*a, **k)

    def tell(self, *a, **k) -> int:
        """Return the cursor position."""
        return self._effective_fileobj.tell(*a, **k)

    def readable(self) -> bool:
        """Is stream readable."""
        return all(getattr(x, "readable", False) for x in self.fileobjs)

    def writable(self) -> bool:
        """Is stream readable."""
        return all(getattr(x, "writable", False) for x in self.fileobjs)

    def seekable(self) -> bool:
        """Is stream readable."""
        return all(getattr(x, "seekable", False) for x in self.fileobjs)


async def _maybe_await(obj: object) -> object:
    if isawaitable(obj):
        return await obj
    else:
        return obj


class async_open(open):
    """Async version of open."""

    async def _close_async(self) -> None:
        # close all the file-like objects that we've created
        for fileobj in reversed(self.fileobjs):
            if isinstance(fileobj, fsspec.asyn.AbstractAsyncStreamedFile):
                await fileobj.close()
            else:
                fileobj.close()
        self.fileobjs = []
        self._is_open = False

    async def _open_async(self) -> None:
        if self._is_open:
            return
        if isinstance(self.fileobj, str):
            await self._fsspec_open_async()
        if self.compression == 'infer' and 'r' in self.mode:
            await self._infer_async()
        self._is_open = True

    async def _fsspec_open_async(self) -> None:
        # Take the last file object (most likely the input fileobj), which
        # must be a path, and open it with `fsspec.open()`.
        uri = stringify_path(self._effective_fileobj)
        # Ensure that input is a string
        if not isinstance(uri, str):
            raise TypeError("Expected a URI, but got:", uri)
        # Set compression option
        opt = dict()
        if not (self.compression == 'infer' and 'r' in self.mode):
            opt['compression'] = self.compression
        # Open with fsspec
        fileobj = await fsopen_async(uri, mode=self.mode, **opt)
        self.fileobjs.append(fileobj)
        # -> returns a fsspec object that's not always fully opened
        #    so open again to be sure.
        if hasattr(fileobj, 'open_async'):
            fileobj = await fileobj.open_async()
            self.fileobjs.append(fileobj)
        elif hasattr(fileobj, 'open'):
            fileobj = fileobj.open()
            self.fileobjs.append(fileobj)

    async def _infer_async(self) -> None:
        # Take the last file object (which at this point must be a byte
        # stream) and infer its compression from its magic bytes.
        fileobj = self._effective_fileobj
        if not hasattr(fileobj, "peek"):
            fileobj = BufferedReader(fileobj)
            self.fileobjs.append(fileobj)
        try:
            if hasattr(fileobj, "peek_async"):
                magic = await fileobj.peek(2)
            else:
                magic = fileobj.peek(2)
        except Exception:
            return
        if magic == b'\x1f\x8b':
            fileobj = GzipFile(fileobj)
            self.fileobjs.append(fileobj)
        if magic == b'BZh':
            fileobj = BZ2File(fileobj)
            self.fileobjs.append(fileobj)

    # ------------------------------------------------------------------
    # open() / IO API
    # ------------------------------------------------------------------

    async def __aenter__(self) -> IO:
        await self._open_async()
        return self._effective_fileobj

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        await self._close_async()

    async def __del__(self) -> None:
        await self._close_async()

    async def open(self) -> IO:
        """Open the file."""
        await self._open_async()
        return self

    async def close(self) -> None:
        """Close the file object."""
        await self._close_async()

    async def read(self, *a, **k) -> bytes | str:
        """Read characters ("t") or byes ("b")."""
        return await _maybe_await(self._effective_fileobj.read(*a, **k))

    async def readline(self, *a, **k) -> bytes | str:
        """Read lines ("t")."""
        return await _maybe_await(self._effective_fileobj.readline(*a, **k))

    async def readinto(self, *a, **k) -> None:
        """Read bytes into an existing buffer."""
        return await _maybe_await(self._effective_fileobj.readinto(*a, **k))

    async def write(self, *a, **k) -> int:
        """Write characters ("t") or bytes ("b")."""
        return await _maybe_await(self._effective_fileobj.write(*a, **k))

    async def writeline(self, *a, **k) -> int:
        """Write lines."""
        return await _maybe_await(self._effective_fileobj.writeline(*a, **k))

    async def peek(self, *a, **k) -> bytes | str:
        """Read the first bytes without moving the cursor."""
        return await _maybe_await(self._effective_fileobj.peek(*a, **k))

    async def seek(self, *a, **k) -> int:
        """Move the cursor."""
        return await _maybe_await(self._effective_fileobj.seek(*a, **k))

    async def tell(self, *a, **k) -> int:
        """Return the cursor position."""
        return await _maybe_await(self._effective_fileobj.tell(*a, **k))

    def readable(self) -> bool:
        """Is stream readable."""
        return all(getattr(x, "readable", False) for x in self.fileobjs)

    def writable(self) -> bool:
        """Is stream readable."""
        return all(getattr(x, "writable", False) for x in self.fileobjs)

    def seekable(self) -> bool:
        """Is stream readable."""
        return all(getattr(x, "seekable", False) for x in self.fileobjs)



# stdlib
import re
from warnings import warn

# externals
import numpy as np

# internals
from .conversions import XYZC, HomogeneousAffineMatrix, Orientation
from .layout import layout_matrix
from .utils import ensure_list


def ensure_homogeneous(mat):
    if mat.shape[-2] != mat.shape[-1]:
        new_shape = list(mat.shape)
        new_shape[-2] += 1
        new_mat = np.zeros_like(mat, shape=new_shape)
        new_mat[..., :-1, :] = mat
        new_mat[..., -1, -1] = 1
        return new_mat
    return mat


def get_voxel_size(mat):
    ndim = mat.shape[-1] - 1
    mat = mat[..., :ndim, :ndim]
    return np.sqrt(np.sum(mat*mat, axis=-2))


def make_vector(x, *args, **kwargs):
    return np.asarray(ensure_list(np.asarray(x).tolist(), *args, **kwargs))


def affine_matmul(a, b):
    """Matrix-matrix product of affine matrices.

    Parameters
    ----------
    a : (..., ndim+1, ndim+1) tensor
        Affine matrix
    b : (..., ndim+1, ndim+1) tensor
        Affine matrix

    Returns
    -------
    affine_times_matrix : (..., ndim+1, ndim+1) tensor

    """
    Za = a[..., :-1, :-1]
    Ta = a[..., :-1, -1:]
    Zb = b[..., :-1, :-1]
    Tb = b[..., :-1, -1:]
    Z = np.matmul(Za, Zb)
    T = np.matmul(Za, Tb) + Ta
    out = np.concatenate((Z, T), axis=-1)
    out = ensure_homogeneous(out)
    return out


def affine_inv(affine):
    """Inverse of an affine matrix.

    If the input matrix is not symmetric with respect to its input
    and output spaces, a pseudo-inverse is returned instead.

    Parameters
    ----------
    affine : (..., ndim[+1], ndim+1) array
        Input affine

    Returns
    -------
    inv_affine : (..., ndim[+1], ndim+1) array

    """
    homogeneous = True
    ndim = affine.shape[-1] - 1
    if affine.shape[-2] != ndim + 1:
        homogeneous = False
        affine = ensure_homogeneous(affine)
    affine = np.linalg.inv(affine)
    if not homogeneous:
        affine = affine[..., :-1, :]
    return affine


# Regex patterns for different value types
patterns = {
    int: r'(\d+)',
    float: r'([\+\-]?\d+\.?\d*(?:[eE][\+\-]?\d+)?)',
    str: r'(.*)\s*$'
}


def read_key(line, key_dict=None):
    """Read one `key = value` line from an LTA file

    Parameters
    ----------
    line : str
    format : type in {int, float, str}

    Returns
    -------
    object or tuple or None

    """
    key_dict = key_dict or dict()

    line = line.split('\r\n')[0]  # remove eol (windows)
    line = line.split('\n')[0]    # remove eol (unix)
    line = line.split('#')[0]     # remove hanging comments
    pattern = r'^\s*(\S+)\s*=\s*(.*)$'
    match = re.match(pattern, line)
    if not match:
        warn(f'cannot read line: "{line}"', RuntimeWarning)
        return None, None

    key = match.group(1)
    value = match.group(2).rstrip()
    if key in key_dict:
        format = key_dict[key]
        if isinstance(format, type):
            pattern = patterns[format]
        else:
            pattern = r'\s*'.join([patterns[fmt] for fmt in format])
        match = re.match(pattern, value)
        if match:
            if isinstance(format, type):
                value = format(match.group(1))
            else:
                value = tuple(fmt(v) for v, fmt in zip(match.groups(), format))
        else:
            warn(f'cannot parse value: "{value}"', RuntimeWarning)
    return key, value


def read_values(line, format):
    """Read one `*values` line from an LTA file

    Parameters
    ----------
    line : str
    format : [sequence of] type
        One of {int, float, str}

    Returns
    -------
    object or tuple or None

    """
    line = line.split('\r\n')[0]  # remove eol (windows)
    line = line.split('\n')[0]    # remove eol (unix)
    line = line.split('#')[0]     # remove hanging comments
    pattern = r'\s*'
    if isinstance(format, type):
        reformat = [patterns[format]]
    else:
        reformat = [patterns[fmt] for fmt in format]
    pattern += r'\s*'.join(reformat)
    value = re.match(pattern, line)
    if value:
        if isinstance(format, type):
            value = format(value.group(1))
        else:
            value = tuple(fmt(v) for v, fmt in zip(value.groups(), format))
    else:
        warn(f'cannot read line: "{line}"', RuntimeWarning)
    return value


def write_key(key, value):
    """Write a `key = value` line in an LTA file.

    Parameters
    ----------
    key : str
    value : int or float or str

    Returns
    -------
    str

    """
    if isinstance(value, (str, int, float)):
        return f'{key:9s} = {value}'
    else:
        return f'{key:9s} = ' + ' '.join([str(v) for v in value])


def write_values(value):
    """Write a `*values` line in an LTA file.

    Parameters
    ----------
    value : [sequence of] int or float or str

    Returns
    -------
    str

    """
    if isinstance(value, (str, int, float)):
        return str(value)
    else:
        return ' '.join([str(v) for v in value])


def fs_to_affine(shape, voxel_size=1., x=None, y=None, z=None, c=0.,
                 source='voxel', dest='ras'):
    """Transform FreeSurfer orientation parameters into an affine matrix.

    The returned matrix is effectively a "<source> to <dest>" transform.

    Parameters
    ----------
    shape : sequence of int
    voxel_size : [sequence of] float, default=1
    x : [sequence of] float, default=[1, 0, 0]
    y: [sequence of] float, default=[0, 1, 0]
    z: [sequence of] float, default=[0, 0, 1]
    c: [sequence of] float, default=0
    source : {'voxel', 'physical', 'ras'}, default='voxel'
    dest : {'voxel', 'physical', 'ras'}, default='ras'

    Returns
    -------
    affine : (4, 4) array

    """
    shape = make_vector(shape)
    voxel_size = make_vector(voxel_size)
    dim = len(list(shape))
    if x is None:
        x = [1, 0, 0]
    if y is None:
        y = [0, 1, 0]
    if z is None:
        z = [0, 0, 1]
    x = make_vector(x, dim)
    y = make_vector(y, dim)
    z = make_vector(z, dim)
    c = make_vector(c, dim)

    shift = shape / 2.
    shift = -shift * voxel_size
    vox2phys = Orientation(shift, voxel_size).affine()
    phys2ras = XYZC(x, y, z, c).affine()

    affines = []
    if source.lower().startswith('vox'):
        affines.append(vox2phys)
        middle_space = 'phys'
    elif source.lower().startswith('phys'):
        if dest.lower().startswith('vox'):
            affines.append(affine_inv(vox2phys))
            middle_space = 'vox'
        else:
            affines.append(phys2ras)
            middle_space = 'ras'
    elif source.lower() == 'ras':
        affines.append(affine_inv(phys2ras))
        middle_space = 'phys'
    else:
        # We need a matrix to switch orientations
        affines.append(layout_matrix(source))
        middle_space = 'ras'

    if dest.lower().startswith('phys'):
        if middle_space == 'vox':
            affines.append(vox2phys)
        elif middle_space == 'ras':
            affines.append(affine_inv(phys2ras))
    elif dest.lower().startswith('vox'):
        if middle_space == 'phys':
            affines.append(affine_inv(vox2phys))
        elif middle_space == 'ras':
            affines.append(affine_inv(phys2ras))
            affines.append(affine_inv(vox2phys))
    elif dest.lower().startswith('ras'):
        if middle_space == 'phys':
            affines.append(phys2ras)
        elif middle_space.lower().startswith('vox'):
            affines.append(vox2phys)
            affines.append(phys2ras)
    else:
        if middle_space == 'phys':
            affines.append(affine_inv(phys2ras))
        elif middle_space == 'vox':
            affines.append(vox2phys)
            affines.append(phys2ras)
        layout = layout_matrix(dest)
        affines.append(affine_inv(layout))

    affine, *affines = affines
    for aff in affines:
        affine = affine_matmul(aff, affine)
    return affine


def affine_to_fs(affine, shape, source='voxel', dest='ras'):
    """Convert an affine matrix into FS parameters (vx/cosine/shift)

    Parameters
    ----------
    affine : (4, 4) array
    shape : (int, int, int)
    source : {'voxel', 'physical', 'ras'}, default='voxel'
    dest : {'voxel', 'physical', 'ras'}, default='ras'

    Returns
    -------
    voxel_size : (float, float, float)
    x : (float, float, float)
    y : (float, float, float)
    z: (float, float, float)
    c : (float, float, float)

    """

    affine = np.asarray(affine)
    vx = get_voxel_size(affine)
    shape = np.asarray(shape)
    source = source.lower()[0]
    dest = dest.lower()[0]

    shift = shape / 2.
    shift = -shift * vx
    vox2phys = Orientation(shift, vx).affine()

    if (source, dest) in (('v', 'p'), ('p', 'v')):
        phys2ras = np.eye(4)

    elif (source, dest) in (('v', 'r'), ('r', 'v')):
        if source == 'r':
            affine = affine_inv(affine)
        phys2vox = affine_inv(vox2phys)
        phys2ras = affine_matmul(affine, phys2vox)

    else:
        assert (source, dest) in (('p', 'r'), ('r', 'p'))
        if source == 'r':
            affine = affine_inv(affine)
        phys2ras = affine

    phys2ras = HomogeneousAffineMatrix(phys2ras)
    return (vx.tolist(), phys2ras.xras().tolist(), phys2ras.yras().tolist(),
            phys2ras.zras().tolist(), phys2ras.cras().tolist())
